
import sys
import os
import argparse
import logging
import yaml
from collections import  Mapping, OrderedDict


# per oglops/yaml_OrderedDict.py
# try to use LibYAML bindings if possible
from yaml import Loader, Dumper
from yaml.representer import SafeRepresenter

_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG

def dict_representer(dumper, data):
	return dumper.represent_dict(data.items())

def dict_constructor(loader, node):
	return OrderedDict(loader.construct_pairs(node))

def represent_none(self, data):
	return self.represent_scalar(u'tag:yaml.org,2002:null', u'')

Dumper.add_representer(OrderedDict, dict_representer)
Dumper.add_representer(type(None), represent_none)

Loader.add_constructor(_mapping_tag, dict_constructor)

Dumper.add_representer(str, SafeRepresenter.represent_str)
Dumper.add_representer(dict, SafeRepresenter.represent_dict)

#Dumper.add_representer(unicode, SafeRepresenter.represent_unicode)

#
# Utility
#

def phex(p, l=0):
	if type(p) == type(""):
		pp = p.encode()
	else:
		pp = p
	hh = ""
	cc = ""
	head = " " * l
	lines = []
	i = 0
	for c in pp:
		hh += "%02x " % c
		if c >= ord(' ') and pp[i] <= ord('~'):
			cc += chr(c)
		else:
			cc += '.'
		if i % 16 == 15:
			lines.append("%s%s %s" % (head, hh, cc))
			hh = ""
			cc = ""
		i += 1
	if cc != "":
		lines.append("%s%-48s %s" % (head, hh, cc))
	return lines


def getargs():
	parser = argparse.ArgumentParser()
	parser.add_argument("-c", "--config",
							help="config file default steamlink.yaml",
							default=None)
	parser.add_argument("-d", "--daemon",
							help="excute as a daemon",
							default=False, action='store_true')
	parser.add_argument("-L", "--loglevel",
							help="set loglevel, default is 'info'",
							default=None)
	parser.add_argument("-l", "--logfile",
							help="file to log to",
							default=None)
	parser.add_argument("-C", "--createconfig",
							help="write default config file and exit",
							default=False, action='store_true')
	parser.add_argument("-p", "--pid-file",
							help="path to pid file when running as daemon",
							default=None)
	parser.add_argument("-T", "--testdata",
							help="generate test data per section 'testdata'",
							default=False)  #, action='store_true')
	parser.add_argument("-v", "--verbose",
							help="print some info",
							default=False, action='store_true')
	parser.add_argument("-X", "--debug",
							help="increase debug level, bumps loglevel to 'debug'",
							default=0, action="count")
	parser.add_argument("-x", "--debugkey",
							help="set of debug keywords, bumps loglevel to ''debug'. Available keywords are: csearch ocache web webupd",
							default=[], action="append")
	parser.add_argument("-V", "--version",
							help="show version and exit",
							default=False, action='store_true')
	return parser.parse_args()


def update(d, u):
	for k, v in u.items():
		if isinstance(v, Mapping):
			d[k] = update(d.get(k, {}), v)
		else:
			d[k] = v
	return d


def loadconfig(default_conf, conf_fname):
	conf = default_conf
	try:
		with open(conf_fname, "r") as fh:
			conf_f = "".join(fh.readlines())
		update(conf, yaml.load(conf_f))
	except FileNotFoundError as e:
		print("note: using default config")
	except Exception as e:
		print("error: config load: %s" % e)
		sys.exit(1)
	return conf


def createconfig(conf_fname, conf):
	if os.path.exists(conf_fname):
		print("error: config file '%s' exists, will NOT overwrite!!" % conf_fname)
		return 1
	with open(conf_fname, 'w') as outfile:
		yaml.dump(conf, outfile,  default_flow_style=False)
	print("note: config written to %s" % (conf_fname))
	return 0


# borrowed from homeassistant
def daemonize() -> None:
	"""Move current process to daemon process."""
	# Create first fork
	pid = os.fork()
	if pid > 0:
		sys.exit(0)

	# Decouple fork
	os.setsid()

	# Create second fork
	pid = os.fork()
	if pid > 0:
		sys.exit(0)

	# redirect standard file descriptors to devnull
	infd = open(os.devnull, 'r')
	outfd = open(os.devnull, 'a+')
	sys.stdout.flush()
	sys.stderr.flush()
	os.dup2(infd.fileno(), sys.stdin.fileno())
	os.dup2(outfd.fileno(), sys.stdout.fileno())
	os.dup2(outfd.fileno(), sys.stderr.fileno())


def check_pid(pid_file: str) -> None:
	"""Check that HA is not already running."""
	# Check pid file
	try:
		pid = int(open(pid_file, 'r').readline())
	except IOError:
		# PID File does not exist
		return

	# If we just restarted, we just found our own pidfile.
	if pid == os.getpid():
		return

	try:
		os.kill(pid, 0)
	except OSError:
		# PID does not exist
		return
	print('Fatal Error: HomeAssistant is already running.')
	sys.exit(1)


def write_pid(pid_file: str) -> None:
	"""Create a PID File."""
	pid = os.getpid()
	try:
		open(pid_file, 'w').write(str(pid))
	except IOError:
		print('Fatal Error: Unable to write pid file {}'.format(pid_file))
		sys.exit(1)


def closefds_osx(min_fd: int, max_fd: int) -> None:
	"""Make sure file descriptors get closed when we restart.

	We cannot call close on guarded fds, and we cannot easily test which fds
	are guarded. But we can set the close-on-exec flag on everything we want to
	get rid of.
	"""
	from fcntl import fcntl, F_GETFD, F_SETFD, FD_CLOEXEC

	for _fd in range(min_fd, max_fd):
		try:
			val = fcntl(_fd, F_GETFD)
			if not val & FD_CLOEXEC:
				fcntl(_fd, F_SETFD, val | FD_CLOEXEC)
		except IOError:
			pass






# coding: utf-8
"""Constants used by Home Assistant components."""
MAJOR_VERSION = 0
MINOR_VERSION = 9
PATCH_VERSION = '0'
__short_version__ = '{}.{}'.format(MAJOR_VERSION, MINOR_VERSION)
__version__ = '{}.{}'.format(__short_version__, PATCH_VERSION)
REQUIRED_PYTHON_VER = (3, 6, 0)


PROJECT_NAME = 'SteamLink'
PROJECT_PACKAGE_NAME = 'steamlink'
PROJECT_LICENSE = 'MIT License'
PROJECT_AUTHOR = 'SteamLink Authors'
PROJECT_COPYRIGHT = ' 2017-2018, {}'.format(PROJECT_AUTHOR)
PROJECT_URL = 'https://steamlink.net/'
PROJECT_EMAIL = 'info+git@steamlink.net'
PROJECT_DESCRIPTION = ('Open-source IoT network framework '
                       'running on Python 3.')
PROJECT_LONG_DESCRIPTION = ('SteamLink is an open-source '
							'IoT networking framework that manages'
							'and runs LoRa radio nodes. ')
PROJECT_CLASSIFIERS = [
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3.6',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content'
]

PROJECT_GITHUB_USERNAME = 'steamlink'
PROJECT_GITHUB_REPOSITORY = 'steamlink'

PYPI_URL = 'https://pypi.python.org/pypi/{}'.format(PROJECT_PACKAGE_NAME)
GITHUB_PATH = '{}/{}'.format(PROJECT_GITHUB_USERNAME,
                             PROJECT_GITHUB_REPOSITORY)
GITHUB_URL = 'https://github.com/{}'.format(GITHUB_PATH)

PLATFORM_FORMAT = '{}.{}'

###

DEFAULT_CONFIG_FILE = "steamlink.yaml"





import aiomqtt
import asyncio
import random
import json
import sys
import os
from hbmqtt.broker import Broker

import logging
logger = logging.getLogger(__name__)

#
# Mqtt
#
class Mqtt:
	def __init__(self, conf, loop = None, as_node=False):
		self.conf = conf
		self.name = "mqtt"
		if loop is None:
			self.loop = asyncio.get_event_loop()
		else:
			self.loop = loop
		self.as_node = as_node

		self.topic_prefix = conf['prefix']
		self.topic_control = conf['control']
		self.topic_data = conf['data']
		self.public_topic_control = conf.get('public_control', None)
		self.public_topic_data = conf.get('public_data', None)
		self.server =   conf['server']
		self.port =     conf['port']
		self.clientid = conf['clientid']
		self.username = conf['username']
		self.password = conf['password']
		self.ssl_certificate = conf.get('ssl_certificate', None)

		self.control_topic_x = "%s/%%s/%s" % (self.topic_prefix, self.topic_control)
		self.control_topic = "%s/+/%s" % (self.topic_prefix, self.topic_control)
		self.data_topic_x = "%s/%%s/%s" % (self.topic_prefix, self.topic_data)
		self.data_topic = "%s/+/%s" % (self.topic_prefix, self.topic_data)

		self.connected = asyncio.Event(loop=loop)
		self.subscribed = asyncio.Event(loop=loop)
		self.disconnected = asyncio.Event(loop=loop)
		self.mq = self.set_mq()
		if self.as_node:
			self.subscription_list = [self.control_topic]
		else:
			self.subscription_list = [self.data_topic]
		if not self.public_topic_control in ['', None]:
			self.public_control_topic = self.public_topic_control % "+"
			self.subscription_list.append(self.public_control_topic)
		else:
			self.public_control_topic=None
		self.running = True

	def set_mq(self):

		mq = aiomqtt.Client(client_id=self.clientid, loop=self.loop)
		mq.loop_start()
#		mq.enable_logger(logger)
		if self.ssl_certificate:
			logger.debug("%s: using cert %s", self.name, self.ssl_certificate)
			try:
				mq.tls_set(self.ssl_certificate)
			except FileNotFoundError as e:
				logger.error("Mqtt: tls_set certificate %s: %s", self.ssl_certificate, e)
				sys.exit(1)
			mq.tls_insecure_set(False)
		if self.username and self.password:
			mq.username_pw_set(self.username, self.password)
		mq.on_connect = self.on_connect
		mq.on_subscribe = self.on_subscribe
		mq.on_message = self.on_message
		mq.on_disconnect = self.on_disconnect
		return mq


	def get_public_control_topic(self):
		return self.public_topic_control

	def set_msg_callback(self, callback):
		if self.as_node:
			logger.debug("set_msg_callback on %s", self.control_topic)
			self.mq.message_callback_add(self.control_topic, callback)
		else:
			logger.debug("set_msg_callback on %s", self.data_topic)
			self.mq.message_callback_add(self.data_topic, callback)

	def set_public_control_callback(self, callback):
		logger.debug("set_public_control_callback on %s", self.public_control_topic)
		self.mq.message_callback_add(self.public_control_topic, callback)


	async def start(self):
		logger.info("%s connecting to %s:%s", self.name, self.server, self.port)

		while True:
			try:
				await self.mq.connect(self.server, self.port, 60)
				break
			except ConnectionRefusedError as e:
				logger.error("Mqtt: connect to %s:%s failed: %s", self.server, self.port, e)
				asyncio.sleep(10)
		await self.wait_connect()


	async def stop(self):
		logger.info("%s done running", self.name)
		self.running = False
		if self.connected.is_set():
			self.mq.disconnect()
			await self.disconnected.wait()

	async def wait_connect(self):
		logger.debug("%s waiting for connect", self.name)
		await self.connected.wait()
		logger.info("%s got connected", self.name)
		for topic in self.subscription_list:
			logger.debug("%s subscribe %s", self.name, topic)
			self.mq.subscribe(topic)
			await self.subscribed.wait()


	def on_connect(self, client, userdata, flags, result):
		logger.info("%s connected %s", self.name, result)
		if result == 0:
			self.disconnected.clear()
			self.connected.set()


	def on_subscribe(self, client, userdata, mid, granted_qos):
		self.subscribed.set()


	def on_disconnect(self, client, userdata, flags):
		self.disconnected.set()
		logger.info("%s: disconnected", self.name)
		self.connected.clear()


	def on_message(self, client, userdata, msg):
		try:
			ddata = json.loads(msg.payload.decode('utf-8'))
		except:
			ddata = msg.payload
		logger.info("%s got %s %s", self.name, msg.topic, ddata)


	def mk_json_msg(self, msg):
		try:
			payload = msg.payload.decode('utf-8')
			jmsg = {'topic': msg.topic, 'payload': payload }
		except:
			jmsg = {'topic': msg.topic, 'raw': msg.payload }

		logger.debug("steamlink msg %s", str(jmsg))
		return jmsg


	def publish(self, firsthop, payload, qos=0, retain=False, sub="control"):
		s = self.control_topic_x if sub == "control" else self.data_topic_x
		topic = s % firsthop
		if 'mqtt' in logging.DBGK:  logger.info("%s publish %s %s", self.name, topic, payload)
		self.mq.publish(topic, payload=payload, qos=qos, retain=retain)


	def public_publish(self, nodename, data, qos=0, retain=False):
		if self.public_topic_data is None:
			return
		topic = self.public_topic_data % nodename
		# logger.debug("%s public publish %s %s", self.name, topic, data)
		self.mq.publish(topic, payload=data, qos=qos, retain=retain)



class Mqtt_Broker(Broker):

	def __init__(self, config=None, loop=None, plugin_namespace=None):
		super().__init__(config, loop, plugin_namespace)
		self.logger.setLevel(logging.WARN)


	async def stop(self):
		await Broker.shutdown(self)




import asyncio
import random
import time

from .steamlink import (
	Steam,
	Mesh,
	Node,
	Packet,
	SL_OP,
)


import logging
logger = logging.getLogger(__name__)

#
# Test
#
class TestData:
	"""" generate  test data in a thread """
	def __init__(self, conf, loop):
		self.name = "TestData"
		self.conf = conf
		self.loop = loop
		self.running = True
#		self.go = asyncio.Event(loop=loop)
		logger.info("starting Test Data")
		self.meshes = {}
		self.nodes = {}
		self.starttime = None


	def stop(self):
		if self.running:
			self.running = False
			logger.debug("%s waiting for shutdown", self.name)


	async def send_test_start(self):
		await asyncio.sleep(self.conf.get('startwait',1))
		await self.send_test(272, "TESTDATA")
		logger.warning("test done")
		return

	async def start(self):
		n_nodes = self.conf.get('nodes',1)
		d_nodes = self.conf.get('del_nodes',0)
		n_meshes = self.conf.get('meshes',1)
		d_meshes = self.conf.get('del_meshes',0)
		n_packets = self.conf.get('packets',1)
		pkt_delay = float(self.conf.get('pkt_delay',0))
		node_delay = float(self.conf.get('node_delay',0))

		logger.info("%s task starting" % self.name)
		await asyncio.sleep(self.conf.get('startwait',1))
		self.starttime = time.time()
		logger.warning("%s test start timing" % self.name)

		for mesh in range(n_meshes):
			logger.debug("creating test mesh %s", mesh)
			self.meshes[mesh] =  registry.find_by_id('Mesh', mesh)
			if self.meshes[mesh] is None:
				self.meshes[mesh] = Mesh(mesh)


		logger.info("%s doing %s nodes", self.name, n_nodes)

		for j in range(n_nodes):
			i = int(random.random() * n_meshes) * 256 + j
			await self.create_node(i)
			await asyncio.sleep(node_delay)

		for x in range(n_packets):
			ii = int(random.random() * n_nodes)
			i = list(self.nodes.keys())[ii]
			await self.create_data(i, "hello from packet %s" % x)
			await asyncio.sleep(pkt_delay)


		for x in range(min(d_nodes, len(self.nodes))):
			i = list(self.nodes.keys())[x]
			logger.info("%s deleting node %s", self.name, i)

			self.nodes[i].delete()
			del self.nodes[i]

		for x in range(min(d_meshes, len(self.meshes))):
			logger.info("%s deleting mesh %s", self.name, x)
			self.meshes[x].delete()
			del self.meshes[x]

		self.running = False
		duration = time.time() - self.starttime
		logger.warning("%s finished, duration %s sec", self.name, int(duration))


	async def create_node(self, i):
		logger.debug("creating test node %s", i)
		self.nodes[i] =  registry.find_by_id('Node', i)
		if self.nodes[i] is None:
			self.nodes[i] = Node(i, nodecfg = None)
			logger.debug("create packet %s", "ON")
		p = Packet(self.nodes[i], sl_op = SL_OP.ON)
		logger.debug("sending ON pkt")
		self.nodes[i].publish_pkt(p, "data")


	async def create_data(self, i, data):
		p = Packet(self.nodes[i], sl_op = SL_OP.DS, payload = "Hello")
		logger.debug("sending DS pkt")
		self.nodes[i].publish_pkt(p, data)


	async def send_test(self, slid, data):
		node = registry.find_by_id('Node', slid)
		if node is None:
			logger.error("no node %s", slid)
			return

		logger.warning("sending %s to node %s", data, node)
		while not node.is_up(): 
			logger.warning("node %s not up, waiting", node)
			await asyncio.sleep(1) 
		rc = node.send_testpacket(data)
		logger.warning("testpkt sent to node %s, code %s", slid, rc)
		



import asyncio
import socketio
import io
import re
import ipaddress
import hashlib
import hmac
import os
import json
import socket
import aiohttp_jinja2
import jinja2
import yaml

from zeroconf import ServiceInfo, Zeroconf
from aiohttp import web
from aiohttp.log import access_logger, web_logger


from .steamlink import (
	Mesh,
	Node,
	Steam,
	Packet,
	add_csearch,
	run_cmd,
	drop_csearch
)
from .linkage import (
	Item,
)

from .const import __version__

import logging
logger = logging.getLogger(__name__)

from yarl import URL

class DisplayConfiguration:

	def __init__(self, file_name):
		self.data = {}
		f = open(file_name)
		self.data = yaml.load(f)
		f.close()

	def row_wise(self):
		rows = {}
		for key in self.data:
			item = self.data[key]
			try:
				if item['row'] not in rows:
					logger.debug("web rowise: itemrow %s ", item['row'])
					rows[item['row']] = []
				rows[item['row']].append(item)
			except KeyError:
				logger.error("DisplayConfiguration: No key found")
		return rows

class NavBar:

	def __init__(self, dir):
		self.navbar_yaml = dir + '/navbar.yaml'
		self.yamls = []
		f = open(self.navbar_yaml)
		self.yamls = yaml.load(f)
		f.close()

#
# WebNameSpace
#
class WebNamespace(socketio.AsyncNamespace):
	def __init__(self, webapp):
		self.webapp = webapp
		self.namespace = self.webapp.namespace

		super().__init__(self.namespace)
		logger.debug("WebNamespace registered for namespace %s", self.namespace)


	def on_connect(self, sid, environ):
		logger.debug("WebNamespace connect %s",str(environ['REMOTE_ADDR']))


	def on_disconnect(self, sid):
		logger.debug("WebNamespace disconnect")
		res = drop_csearch(self, sid, {})


#	async def on_need_log(self, sid, data):
#		logger.debug("WebNamespace need_log %s", data)
#	#	await self.emit('my_response', {'data': data['data']} ) #, room=sid, namespace=self.namespace)
#		node = data.get('id',None)
#		if  not node in Node.name_idx:
#			return "NAK"
#		try:
#			r = Node.name_idx[node].console_pkt_log(data['key'], int(data['count']))
#		except:
#			return "NAK"
#		return "ACK"


	async def on_startstream(self, sid, message):
		logger.debug("WebNamespace on_startstream --> %s", message)
		try:
			res = add_csearch(self, sid, message)
		except KeyError as e:
			msg = '%s field missing in request' % e
			logger.warning(msg)
			raise
			return {'error': msg }
		except TypeError as e:
			msg = '%s, probably incorrect value for start_key' % e
			logger.warning(msg)
			raise
			return {'error': msg }

		logger.debug("WebNamespace on_startstream <-- %s", res)
		return res


	async def on_leave(self, sid, message):

		logger.debug("WebNamespace on_leave %s", message)

		res = drop_csearch(self, sid, message)
		return { 'success': True }


	# dict with: cmd, node_name, data
	# cmds are: DN SC
	async def on_cmd(self, sid, message):
		logger.debug("WebNamespace on_cmd --> %s", message)
		try:
			ret = run_cmd(self, sid, message)
			res = { "Success": ret }
		except  Exception as e:
			res = { "Success": False, "Error": "cmd failed: %s" % e }

		logger.debug("WebNamespace on_cmd <-- %s", res)
		return res

#
# WebApp
#
class WebApp(object):

	def __init__(self, namespace, sio, conf, loop = None):
		self.name = "WebApp"
		self.conf = conf
		self.minupdinterval = conf['minupdinterval']
		self.sio = sio
		self.namespace = namespace
		self.loop = loop
		self.con_upd_q = asyncio.Queue(loop=self.loop)
		self.app = web.Application()
		self.app._set_loop(self.loop)
		self.sio.attach(self.app)
		self.app['websockets'] = []
		self.app.router.add_get('/config.json', self.config_json)

		self.libdir = os.path.dirname(os.path.abspath(__file__))
		self.static_dir = self.libdir+'/html/static'
		self.templates_dir = self.libdir+'/html/templates'
		self.app.router.add_route('GET', '/', self.route_handler)
		self.app.router.add_route('GET', '/favicon.ico', self.favicon_handler)
		self.app.router.add_route('GET', '/ghwh', self.ghwh_handler)
		self.app.router.add_route('POST', '/ghwh', self.ghwh_handler)
		self.app.router.add_route('GET', '/{file_name}', self.route_handler)


		self.app.router.add_static('/static', self.static_dir)
		self.app.on_cleanup.append(self.web_on_cleanup)
		self.app.on_shutdown.append(self.web_on_shutdown)
		self.backlog = 128

		aiohttp_jinja2.setup(self.app, loader=jinja2.FileSystemLoader(self.templates_dir))

		self.shutdown_timeout = self.conf['shutdown_timeout']
	#	self.api_password = conf['api_password']
		self.ssl_certificate = conf['ssl_certificate']
		self.ssl_key = conf['ssl_key']
		self.ssl_context = None
		self.access_log_format = 'XXX %a %t "%r" %s %b "%{Referer}i" "%{User-Agent}i"'
		self.access_log = access_logger
		access_logger.setLevel(logging.WARN)	# N.B. set in config

		self.host = conf['host']
		self.port = conf['port']
	#	self._handler = None
	#	self.server = None

	def __getstate__(self):
		return {'MyClass': 'WebApp'}

	async def qstart(self):
		logger.info("%s starting q handler", self.name)
		self.con_upd_res = await self.console_update_loop()


	async def start(self):
		logger.info("%s starting, server %s port %s", self.name, self.host,  self.port)
		self.sio.register_namespace(WebNamespace(self))
		self.runner = web.AppRunner(self.app)
		await self.runner.setup()
		self.site = web.TCPSite(self.runner, 'localhost', self.port)
		await self.site.start()

		logger.debug("%s: app started", self.name)

		scheme = 'https' if self.ssl_context else 'http'


		my_ip_address = socket.gethostbyname(socket.gethostname())
		desc = {'version': __version__}
		self.zeroconf_info = ServiceInfo("_http._tcp.local.",
					"SteamLink Store._http._tcp.local.",
					socket.inet_aton(my_ip_address), self.port, 0, 0,
					desc, "steamlink.local.")
		self.zeroconf = Zeroconf()
		self.zeroconf.register_service(self.zeroconf_info)

		make_handler_kwargs = dict()
		make_handler_kwargs['access_log_format'] = self.access_log_format
		make_handler_kwargs['access_log'] = self.access_log

		self.handler = self.app.make_handler(loop=self.loop, **make_handler_kwargs)

		self.server = await self.loop.create_server(
						self.handler, self.host, self.port,
						ssl=self.ssl_context,
						backlog=self.backlog)

	def stop(self):
		logger.info("%s done running", self.name)
		self.zeroconf.unregister_service(self.zeroconf_info)
		self.server.close()
#		somethong int this list may take 20 seconds to shutdown gracefully
		self.loop.run_until_complete(self.server.wait_closed())
		self.loop.run_until_complete(self.app.shutdown())
#		self.loop.run_until_complete(self.handler.shutdown(self.shutdown_timeout))
#		self.loop.run_until_complete(self.runner.cleanup())


	async def config_json(self, request):
		rj = json.dumps(self.conf)
		return web.Response(text=rj, content_type='application/json')


	def web_on_cleanup(self, app):
		logger.info("web closing down")


	async def web_on_shutdown(self, app):
		for ws in self.app['websockets']:
			await ws.close(code=WSCloseCode.GOING_AWAY, message='Server shutdown')


	def queue_itemlist_update(self, csitems, force):
		if 'webupd' in logging.DBGK: logger.debug("queue_itemlist_update for %s item %s", csitem.csearch.search_id, csitem.item)
		asyncio.ensure_future(self.con_upd_q.put([csitems, force]), loop=self.loop)



	def queue_item_update(self, csitem, force):
		if 'webupd' in logging.DBGK: logger.debug("queue_item_update for %s item %s", csitem.csearch.search_id, csitem.item)
		asyncio.ensure_future(self.con_upd_q.put([csitem, force]), loop=self.loop)


	async def console_update_loop(self):
		logger.info("%s q handler", self.name)
		while True:
			upd_csitem, upd_force =  await self.con_upd_q.get()
			if 'webupd' in logging.DBGK: logger.debug("console_update_loop %s force %s", upd_csitem, upd_force)
			if upd_csitem is None:
				break
			if type(upd_csitem) == type([]):
				data = []
				for cs_item in upd_csitem:
					data.append(cs_item.console_update(upd_force))
				upd_csitem = upd_csitem[0]
			else:
				data = upd_csitem.console_update(upd_force)
			tag = upd_csitem.csearch.csearchkey.stream_tag
			room = upd_csitem.csearch.search_id
			if 'webupd' in logging.DBGK: logger.debug("emit_loop event: %s room:%s data: %s", upd_csitem.csearch.csearchkey.stream_tag,  upd_csitem.csearch.search_id, data)
			await self.sio.emit(tag,
					data = data,
					namespace = self.namespace,
					room = room)
			upd_csitem.update_sent()
			self.con_upd_q.task_done()

		logger.debug("console_update_loop done")


	async def console_alert(self, lvl, smsg):
		if len(smsg) > 110:
			msg = smsg[:110]+"..."
		else:
			msg = smsg
		alert = {'lvl': lvl, 'msg': msg }
		await self.sio.emit('alert', alert, namespace = self.namespace)


	def send_console_alert(self, lvl, msg):
		asyncio.ensure_future(self.console_alert(lvl, msg), loop=self.loop)


	async def favicon_handler(self, request):
		return None


	async def ghwh_handler(self, request):
		""" GitHub WebHook handler """

		if request.method == 'GET':
			return web.Response(text='OK')
		elif request.method == 'POST':
			# Store the IP address of the requester
			request_ip = request.remote

#			# If VALIDATE_SOURCEIP is set to false, do not validate source IP
#			if self.conf.get('repo_validate_sourceip', False):
#
#				# If GHE_ADDRESS is specified, use it as the hook_blocks.
#				if self.conf.get('repo_ghe_address', None):
#					hook_blocks = self.conf.get('repo_ghe_address')
#				# Otherwise get the hook address blocks from the API.
#				else:
#					hook_blocks = requests.get('https://api.github.com/meta').json()['hooks']
#
#				# Check if the POST request is from github.com or GHE
#				for block in hook_blocks:
#					if ipaddress.ip_address(request_ip) in ipaddress.ip_network(block):
#						break  # the remote_addr is within the network range of github.
#				else:
#					if str(request_ip) != '127.0.0.1':
#						return web.Response(text='invalid IP', status=403)

			request_data = await request.text()
			payload = json.loads(request_data)

			if request.headers.get('X-GitHub-Event') == "ping":
				logger.debug("ghwh ping data %s", payload)
				return web.json_response({'msg': 'Hi!'})
			if request.headers.get('X-GitHub-Event') != "push":
				return web.json_response({'msg': "wrong event type"})

			full_name = "%s/%s" % (self.conf['repo_owner'], self.conf['repo_name'])
			if full_name != payload['repository']['full_name']:
				logger.debug("ghwh push repo not ours %s", payload['repository']['full_name'])
			else:
				logger.debug("ghwh push for repo %s", full_name)

				# Check if POST request signature is valid
				key = self.conf.get('repo_key', None)
				if key is None:
					logger.debug("ghwh no key")
				else:
					signature = request.headers.get('X-Hub-Signature')
					if signature[:5] == "sha1=":
						signature = signature[5:]
					logger.debug("sig %s %s %s %s", type(key), key, type(signature), signature)
					if type(key) != type(b''):
						key = key.encode()
					if type(request_data) !=  type(b''):
						request_data = request_data.encode()
					mac = hmac.new(key, msg=request_data, digestmod=hashlib.sha1)
					if not hmac.compare_digest(mac.hexdigest(), signature):
						logger.warning("ghwh auth failed mac %s sig %s", mac.hexdigest(), signature)
						return web.Response(text='auth fail', status=403)

					ci_cmd = self.conf.get("repo_ci_command", None)
					if ci_cmd is not None:
						logger.warning("ghwh upgrading running version")
						os.system(ci_cmd + "&")

			return web.Response(text='OK')



	async def route_handler(self, request):
		nav = NavBar(self.templates_dir)

		if request.rel_url.path == '/':
			file_name = 'index'
		else:
			file_name = request.match_info['file_name']
			# file_name = str(request.rel_url.path).rstrip('/')
		logger.debug("web route_handler: filename: %s, request.rel_url %s", file_name, request.rel_url)
		try:
			dc = DisplayConfiguration(self.templates_dir + '/' + file_name + '.yaml')
		except FileNotFoundError as e:
			return web.Response(text="No such file: %s" % file_name, status=404)
		for qk in request.query:
			if 'web' in logging.DBGK: logger.debug("web route_handler: Query' : %s =%s", qk, request.query[qk])
			try:
				partial, key = qk.split('.', 1)
			except:
				if 'web' in logging.DBGK: logger.debug("web route_handler: Query key with no '.' : %s", qk)
				continue
			if not partial in dc.data:
				if 'web' in logging.DBGK: logger.debug("web route_handler: partial not in yaml : %s", partial)
				continue
			if 'web' in logging.DBGK: logger.debug("web route_handler: dc.data[partial] : %s", dc.data[partial])
			i = request.query[qk]
			try:
				i = int(i)
			except:
				pass

			if (key == "restrict_by"):
				dc.data[partial][key][0]['value'] = i
			else:
				dc.data[partial][key] = i

		if 'web' in logging.DBGK: logger.debug("webapp handler %s", dc.data)
		context = { 'context' : dc.row_wise(), 'navbar' : nav.yamls }

		if not (os.path.isfile(self.templates_dir + '/'+ file_name + '.html')):
			file_name = 'index'
		response = aiohttp_jinja2.render_template(file_name + '.html', request, context)
#		response.headers['Content-Language'] = 'en'
		return response




import asyncio
from asyncio import Queue
import shelve
import os
import re
import sys
import socket
import time
from collections import  Mapping, OrderedDict

import logging
logger = logging.getLogger()

# Globals, initialized by attach
_WEBAPP = None
_DB = None

def Attach(webapp, db):
	global _WEBAPP, _DB
	if _WEBAPP is not None:
		logger.error("Linkage: Attach already done")
		return
	_WEBAPP = webapp
	_DB = db

	if _WEBAPP is None:
		webapp_name = "-"
	else:
		webapp_name = _WEBAPP.name
	if _DB is None:
		db_name = "-"
	else:
		db_name = _DB.name

	logger.debug("linkage: Attached apps '%s, %s'", webapp_name, db_name)



def check_restrictions( restrict_by, item):
	if len(restrict_by) == 0:
		return True
	res = True
	for restrict in restrict_by:
		field =  restrict['field_name'] 
		op =  restrict['op'] 
		value =  restrict['value'] 
		ex = "item['%s'] %s %s" % (field, op, repr(value))
		res = res and eval(ex)
	return res


import yaml
from yaml import Loader, Dumper

class CSearchKey:
	def __init__(self, table_name, key_field, start_key, start_item_number,
					    count, stream_tag="NoTag", end_key = None, restrict_by = []):

		self.table_name = table_name
		self.key_field = key_field
		if key_field == 'ts' and start_key is not None:					# to help javascript...
			self.start_key = float(start_key)
		else:
			self.start_key = start_key
		self.start_item_number = start_item_number
		self.end_key = end_key
		self.count = count
		self.stream_tag = stream_tag

		self.at_start = False
		self.at_end = False
		self.total_item_count = 0

		self.restrict_by = restrict_by
		self.search_id =  self.__repr__()	#used to index CSearches 


	def __repr__(self):
		return  "%s(%s:%s:%s)_%s_%s_%s" %\
			(self.table_name, self.key_field, self.start_key, self.end_key,  \
			self.restrict_by, self.start_item_number, self.stream_tag)

	def __str__(self):
		at_start = "S" if self.at_start else "-"
		at_end = "E" if self.at_end else "-"
		return "CS: %s %s(%s->%s) -%s- tot %s  cnt %s %s%s" % \
			(self.table_name, self.key_field, self.start_key, self.end_key, 
			self.restrict_by, self.total_item_count, self.count, at_start, at_end)
#
# CSearch 
#
class CSearch:
	def __init__(self,  webnamespace, table, csearchkey):

		self.webnamespace = webnamespace
		self.csearchkey = csearchkey
		self.search_id = csearchkey.search_id
		self.clients = {}		# users who registed, sid, stream_tag
		self.cs_items = OrderedDict()		# current list of items in sarch, key is key_field

		self.table = table
		if csearchkey.key_field is None:
			csearchkey.key_field = self.table.keyfield
		for item in self.table.get_range(self.csearchkey):
			self.add_item(item)
		if 'csearch' in logging.DBGK: logger.debug("CSearch csearch key: %s", str(csearchkey))


	def __str__(self):
		return "CSearch[%s]" % self.search_id


	def add_item(self, item):
		if 'csearch' in logging.DBGK: logging.debug("CSearch '%s' add_item %s in key %s", self.search_id, item, self.csearchkey.key_field)
		self.cs_items[item.__dict__[self.csearchkey.key_field]] = CSearchItem(self, item)


	def drop_item(self, item):
		if 'csearch' in logging.DBGK: logger.debug("CSearch '%s' drop_item %s", self.search_id, item)
		self.cs_items[item.__dict__[self.csearchkey.key_field]].deleted = True
#?		del self.cs_items[item.__dict__[self.csearchkey.key_field]]


	def add_sid(self, sid):
		if 'csearch' in logging.DBGK: logger.debug("CSearch '%s' add_sid %s csearchkey %s", self.search_id, sid, self.csearchkey)
		self.clients[sid] = self.csearchkey.stream_tag
		self.webnamespace.enter_room(sid, self.search_id, \
			 self.webnamespace.namespace)
		return self.csearchkey


	def drop_sid(self, sid):
		if sid in self.clients:
			if 'csearch' in logging.DBGK: logger.debug("CSearch '%s' drop_sid %s", self.search_id, sid)
			del self.clients[sid]
			self.webnamespace.leave_room(sid, self.search_id, \
				 self.webnamespace.namespace)


	def num_clients(self):
		return len(self.clients)


	def check_csearch(self, op, item, force = False):
		""" check if item matches any registered searches, 
			schedule update for all search clients it matched 
			op is 'ins', 'upd' or 'del'
			note that item is updated or inserted in db before check_csearch
			but the delete will happen after
		"""
# find csitem for item
		if 'csearch' in logging.DBGK: logger.debug("check_csearch (CSearch) %s force=%s op=%s item=%s",\
				 self, force, op, item)
		item_search_key = item.__dict__[self.csearchkey.key_field]
		push = False
		go = check_restrictions(self.csearchkey.restrict_by, item.__dict__)
		if not go:
			return

		if not item_search_key in self.cs_items:
			if op in ['ins']:
				if  self.csearchkey.at_end \
					and  item_search_key > self.csearchkey.end_key:
					push = True
					self.csearchkey.end_key = item_search_key
					self.add_item(item)
				elif not push and op in ['ins' ,'upd'] \
					and self.csearchkey.at_end \
					and  item_search_key < self.csearchkey.start_key:
					push = True
					self.csearchkey.start_key = item_search_key
					self.add_item(item)
		elif item_search_key >= self.csearchkey.start_key and \
				item_search_key <= self.csearchkey.end_key:
			if op == 'upd':
				pass
			elif op == 'ins':
				self.add_item(item)
			elif op == 'del':
				self.drop_item(item)

			push = True

		if push or force:
			if 'csearch' in logging.DBGK: logger.debug("check_csearch (push) %s force=%s", self, force)
			try:
				self.cs_items[item_search_key].push_update(force)
			except KeyError as e:
				logger.error("check_csearch (push) keyerror: %s", e)
		return 


	def force_update(self, sid):
		if not sid in self.clients:
			return
		if 'csearch' in logging.DBGK: logger.debug("force_update ")


		cs_list = []
		for cs in self.cs_items:
			cs_list.append(self.cs_items[cs])
		if len(cs_list) > 0:
			_WEBAPP.queue_itemlist_update(cs_list, True)


#
# CSearchItem
#
class CSearchItem:
	def __init__(self, csearch, item):
		self.csearch = csearch
		self.item = item
		# stash for when item is deleted
#		self.itemname = "%s" % item.name
		self.deleted = False

		self.last_update = 0		# csearchitem's last update time stamp
		self.future_update = False
		self.upd_in_progress = False


#	def __getstate__(self):
#		r = self.__dict__.copy()
#		r['csearch'] = self.csearch.csearchkey 
#		r['item'] = self.item.key
#		return r

	def __repr__(self):
		return str("CSI %s %s" % (self.csearch.csearchkey, self.item))

	def console_update(self, force):
		if 'csearch' in logging.DBGK: logger.debug("console_update %s %s", self, force)
		data_to_emit = self.item.gen_console_data()
		if self.deleted:
			data_to_emit['_del_key'] = 1

		if 'csearch' in logging.DBGK: logger.debug("console_update CSearch %s Item %s Data %s", self.csearch, self.item, str(data_to_emit)[:32]+"...")
		return data_to_emit


	async def schedule_future_update(self, wait):
		await asyncio.sleep(wait)
		self.future_update = False
		self.push_update(False)


	def push_update(self, force):
		if self.upd_in_progress or _WEBAPP is None:
			if 'csearch' in logging.DBGK: logger.debug("push_update upd_in_progress %s", self)
			return
		next_update = (self.last_update + _WEBAPP.minupdinterval) - _WEBAPP.loop.time()
		if 'csearch' in logging.DBGK: logger.debug("push_update %s %s next: %s", self, force, next_update)
		if not force and next_update > 0:
			if not self.future_update:
				asyncio.ensure_future(self.schedule_future_update(next_update))
				self.future_update = True
			return

		if 'csearch' in logging.DBGK: logger.debug("push_update %s %s next: %s", self, force, next_update)
		self.upd_in_progress = True
		_WEBAPP.queue_item_update(self, force)


	def update_sent(self):
		self.upd_in_progress = False
		self.last_update = _WEBAPP.loop.time()



#
# OCache
#
class OCache(dict):
	def __init__(self, tablename, max_entries = 1000):
		self.tablename = tablename
		self.max_entries = max_entries
		self.ts = {}
		super().__init__()
		self.warned = False
		self.gets = 0
		self.sets = 0
		self.hits = 0
		self.misses = 0
		self.replaces = 0
		if 'ocache' in logging.DBGK: logger.debug("OCache %s created", self.tablename)

	def __del__(self):
		if 'ocache' in logging.DBGK: logger.debug("OCache %s destroyed", self.tablename)
		return super().__del__()

	def __getitem__(self, key):
		self.gets += 1
		self.ts[key] = time.time()
		if 'ocache' in logging.DBGK: logger.debug("OCache %s __getitem %s", self.tablename, key)
		return super().__getitem__(key)


	def __delitem__(self, key):
		if 'ocache' in logging.DBGK: logger.debug("OCache %s __delitem %s", self.tablename, key)
		try:
			return super().__delitem__(key)
		except:
			return


	def has(self, key):
		if key in self:
			self.hits += 1
			return True
		self.misses += 1
		return False

	def __setitem__(self, key, value):
		if key in self:
			self.replaces += 1
		super().__setitem__(key, value)
		self.sets += 1
		self.ts[key] = time.time()
		if 'ocache' in logging.DBGK: logger.debug("OCache %s __setitem__ %s %s at %s", self.tablename, key, value, self.ts[key])
		if len(self) > self.max_entries:
			self.clean()
		else:
			self.warned = False


	def clean(self):
		if 'ocache' in logging.DBGK: logger.debug("OCache %s clean!", self.tablename)
#		to_prune =  len(self) - self.max_entries
		to_prune =  int(self.max_entries / 10)
		for p in range(to_prune):
			ts = time.time()
			d = None
			for i in self.keys():
				if sys.getrefcount(super().__getitem__(i)) == 2:	# N.B sys.getrefcount and  self[i]
					if ts > self.ts[i]:
						d = i
						ts = self.ts[i]
			if d is None:
				break
			del self[d]
			del self.ts[d]

		if len(self) >= self.max_entries:
				self.warned = True
				logger.debug("cache overcommited: "+self.status())

		if 'ocache' in logging.DBGK: logger.debug("OCache cleaned: %s", self.status())
		return 

	def status(self):
		in_use = 0
		occupied = len(self)
		for i in self.keys():
			if sys.getrefcount(super().__getitem__(i)) != 2:
				in_use += 1
		return "'%s' %s in use, %s of %s ocupied (%s%%) m %s h %s s %s g %s r %s" % \
			(self.tablename, in_use, occupied, self.max_entries, (100.0 * occupied / self.max_entries),
				self.misses, self.hits, self.sets, self.gets, self.replaces)


#
# Table
#
class Table:
	tables = {}

	""" provide an index over items """
	index = None
	def __init__(self, itemclass, keyfield):
		self.itemclass = itemclass
		self.keyfield = keyfield
		self.csearches = {}
		self.sid_stream_tags = {}
		Table.tables[self.itemclass.__name__] = self


	def add_csearch(self, webnamespace, csearchkey, sid):
		""" add a cserarch for sid to this table """

		self.drop_stream_tag_from_csearch(sid, csearchkey.stream_tag)
		srch_id = csearchkey.search_id
		if not srch_id in self.csearches:
			if 'webupd' in logging.DBGK: logger.debug("table add_csearch '%s' new: %s", srch_id, csearchkey)
			self.csearches[srch_id] = CSearch(webnamespace, self, csearchkey)
		else:
			if 'webupd' in logging.DBGK: logger.debug("table add_csearch search_id '%s' from cache", srch_id)
		csearchkey = self.csearches[srch_id].add_sid(sid)
		return csearchkey


	def drop_stream_tag_from_csearch(self, sid, stream_tag):
		""" delete sid from any csearches with the same stream_tag """
		del_client = []
		for cs in self.csearches:
			if sid in self.csearches[cs].clients \
				and stream_tag == self.csearches[cs].csearchkey.stream_tag:
				del_client.append(cs)

		for dc in  del_client:
			if 'webupd' in logging.DBGK: logger.debug("table drop_stream_tag_from_csearch t %s", dc)
			del self.csearches[dc].clients[sid]
			if len(self.csearches[dc].clients) == 0:
				del self.csearches[dc]


	def drop_sid_from_csearch(self, sid):
		""" delete sid from all csearches """
		del_list = []
		for cs in self.csearches:
			self.csearches[cs].drop_sid(sid)
			if 'webupd' in logging.DBGK: logger.debug("table drop_sid_from_csearch '%s'", sid)
			if self.csearches[cs].num_clients() == 0:
				del_list.append(cs)
		for cs in del_list:
			if 'webupd' in logging.DBGK: logger.debug("table drop_sid_from_csearch csearch empty! '%s'", cs)
			del self.csearches[cs]

	def register(self, item):
		""" either load item from backing store, if it exists, or store it there """
		pass


	def find(self, key, keyfield = None):
		pass

	def find_one(self, key, keyfield = None):
		pass

	def insert(self, item):
		self.check_csearch('ins', item, force=False)

	def db_update(self, item, force):
		self.check_csearch('upd', item, force)

	def delete(self, item):
		self.check_csearch('del', item, force=False)

	def check_csearch(self, op, item, force):
		for cs in self.csearches:
			if 'webupd' in logging.DBGK: logger.debug("check_csearch (Table) %s force=%s", cs, force)
			self.csearches[cs].check_csearch(op, item, force)
		 
#
# DbTable
#
class DbTable(Table):
	""" database based Table """

	def __init__(self, itemclass, keyfield, tablename):
		self.tablename = tablename
		self.cache = OCache(tablename, 1000)
		self.dbtable = _DB.table(self.tablename)
		super().__init__(itemclass, keyfield)


	def register(self, item):
		""" backload item from db if it exists, otherwise insert in db """
		if logging.DBG > 2: logger.debug("register %s %s", self.tablename, item)
		item_dict = self.dbtable.get(self.keyfield, '==',  item.__dict__[self.keyfield])
		if item_dict is None:
			item._wascreated = True
			self.insert(item)
		else:
			item._wascreated = False
			item.load(item_dict)
			self.cache[item.__dict__[self.keyfield]] = item
		super().register(item)
		return item._wascreated


	def make_item_from_dict(self, item_dict):
		if item_dict is None:
			return None
		key = item_dict[self.keyfield]
		if self.cache.has(key):
			item = self.cache[key]
		else:
			item = self.itemclass()
			item.load(item_dict)
			self.cache[key] = item
		return item


	def get_range(self, csk):
		""" get a range of items, defined by a CSearch """
		for item_dict in self.dbtable.get_range(csk):
			yield self.make_item_from_dict(item_dict)


	def find(self, key, keyfield = None):
		if keyfield is None:
			keyfield = self.keyfield
		for item_dict in self.dbtable.search(keyfield, '==', key):
			yield self.make_item_from_dict(item_dict)


	def find_one(self, key, keyfield = None):
		""" find exactly one item with keyfield key """
		if keyfield is None:
			keyfield = self.keyfield
		if keyfield == self.keyfield:	# i.e. native key
			if self.cache.has(key):
				return self.cache[key]
		item_dict = self.dbtable.get(keyfield, '==',  key)
		if logging.DBG > 1: logger.debug("find_one %s %s: %s", self.tablename, key,  res)
		return self.make_item_from_dict(item_dict)

#?		logger.error("multiple items with '%s' = %s in table %s", keyfield, key, self.tablename)


	def db_update(self, item, force=False):
		if 'webupd' in logging.DBGK: logger.debug("db_update (DBTable) %s force=%s", self, force)
		self.dbtable.db_update(item.save(), self.keyfield, item.__dict__[self.keyfield])
		self.cache[item.__dict__[self.keyfield]] = item
		super().db_update(item, force)


	def insert(self, item):
		if logging.DBG > 2: logger.debug("Table insert %s  %s", item.save(), self.keyfield)
		self.dbtable.upsert(item.save(), self.keyfield)
		self.cache[item.__dict__[self.keyfield]] = item
		super().insert(item)


	def delete(self, item):
		logger.debug("DbTable delete %s", item)
		del self.cache[item.__dict__[self.keyfield]]
		self.dbtable.delete(self.keyfield, item.__dict__[self.keyfield])
		super().delete(item)


	def __len__(self):
		return len(self.dbtable)

	def __str__(self):
		return "Dbindex(%s)%s" % (self.itemclass.__name__, len(self.dbtable))


#
# DictTable
#
class DictTable(Table):
	""" dict based Table """

	def __init__(self, itemclass,  keyfield, index):
		self.tablename = itemclass.__name__
		self.index = index
		super().__init__(itemclass, keyfield)


	def register(self, item):
		logger.debug("DictTable register %s %s", item.__dict__[self.keyfield], item)
		self.index[item.__dict__[self.keyfield]] = item


	def get_range(self, csk):
		if logging.DBG > 1: logger.debug("get_range csk %s", str(csk))
		field = csk.key_field
		startv = csk.start_key
		endv = csk.end_key
		count = csk.count

		csk.total_item_count = 0

		tab = self.index
		if len(tab) == 0:
			if logging.DBG > 1: logger.debug("get_range table empty")
			csk.count = 0
			return {}

		udict = {}
		for t in tab:
			logger.debug("DictTable get_range %s %s", field, t)

			udict[tab[t].__dict__[field]] = tab[t]
		fullsdict = sorted(udict)
		if logging.DBG > 1: logger.debug("get_range table %s items", len(fullsdict))

		if len(csk.restrict_by) == 0:
			sdict = fullsdict
		else:
			sdict = []
			for r in fullsdict:
				if check_restrictions(csk.restrict_by, udict[r]):
					sdict.append(r)

		if len(sdict) == 0:
			if logging.DBG > 1: logger.debug("get_range table empty after destrict")
			return {}

		if startv in [None]:
			if csk.start_item_number < 0:
				sidx = max(0, len(sdict) + csk.start_item_number)
			else:
				sidx = min(csk.start_item_number, len(sdict)-1)
			startv = sdict[sidx]
		else:
			sidx = None
			for idx in range(len(sdict)):
				if sdict[idx] >=  startv:
					sidx = idx
					startv = sdict[sidx]
					break
			if sidx is None:
				if logging.DBG > 1: logger.debug("get_range no start key found")
				return {}
		if endv in [None]:
			eidx = min(sidx + count-1, len(sdict)-1)
			endv = sdict[eidx]
		else:
			eidx = None
			for idx in range(sidx, len(sdict)):
				if sdict[idx] <= endv:
					endv = sdict[idx]
					eidx = idx
					break
			if eidx is None:
				if logging.DBG > 1: logger.debug("get_range no end key found")
				return {}
			count = eidx - sidx + 1
		res = {}
		for idx in range(sidx, eidx+1):
			res[sdict[idx]]  = udict[sdict[idx]]

		csk.start_key = sdict[sidx]
		csk.end_key = sdict[eidx]
		csk.start_item_number = sidx
		csk.total_item_count = len(sdict)
		csk.at_start = csk.start_key == sdict[0]
		csk.at_end = csk.end_key == sdict[-1]

		for r in res:
			yield res[r]


	def find(self, key, keyfield = None):
		if keyfield is None or keyfield == self.keyfield:	# native key searc
			try:
				return self.index[key]
			except:
				return None
		res = []
		for k in self.index.keys():
			if self.index[k][keyfield] == value:
				res.append(self.index[k])
		return res

	def db_update(self, item, force=False):
		if 'webupd' in logging.DBGK: logger.debug("db_update (DictTable)  %s force=%s", self, force)
		super().db_update(item, force)


	def insert(self, item):
		if logging.DBG > 2: logger.debug("DictTable  insert %s", item.save())
		pass


	def delete(self, item):
		del item
		pass


	def __len__(self):
		return len(self.index)


	def __str__(self):
		return "DictTable(%s)%s" %(self.itemclass.__name__, len(self.index))

#
# BaseItem
#
class BaseItem:
	def __init__(self, key):
		self._key = key		# convenient shortcut for  self.__dict__[self._table.keyfield]


	def __del__(self):
		if logging.DBG > 2: logger.debug("BaseItem: __del__ %s", self)


	def __str__(self):
		try:
			return "%s (%s)" % (self._itype, self._key)
		except AttributeError:
			return "SomeBaseItem"


#
# Item
#
class Item(BaseItem):
	_table = None

	# three ways Items are instanciated:
	# 1. Normal:   _load = None and key != None
	# 2. Load:	   _load != None 
	# 3. Empty:    _load = None and key = None
	def __init__(self, key, _load=None):
		super().__init__(key)
		self._table = self.__class__._table
		self._itype = self.__class__.__name__
		if _load is not None:
			self.load(_load)
			if logging.DBG >= 1: logger.debug("Item loaded")
		elif key is not None: 		# skip key-less items, they will come in via load()
			self._table.register(self)
			if logging.DBG >= 1: logger.debug("Item created and loaded: %s", self)
		else:
			if logging.DBG >= 1: logger.debug("Item created null: %s", self)


	def load(self, data):	#N.B.
		""" load class variable from provided data """
		if logging.DBG > 2: logger.debug("%s loading data: %s", self.itype, data)
		for k in data:
			self.__dict__[k] = data[k]
		self._key = self.__dict__[self._table.keyfield]
		logger.debug("Item loaded: %s", self)


	def save(self, withvirtual=False):
		""" return dict of all non-private class variables """
		r = {}
		for k in self.__dict__:
			if k[0] == '_':
				continue
			r[k] = self.__dict__[k]
		return r

	def insert(self):
		if logging.DBG > 2: logger.debug("Item  insert %s %s", self._table.tablename, self.save())
		self._table.insert(self)

	def update(self, force=False):
		if 'webupd' in logging.DBGK: logger.debug("Item update %s force=%s", self, force)
		self._table.db_update(self, force)

	def delete(self):
		self._table.delete(self)

	def gen_console_data(self):
		return self.save(withvirtual=True)

#
# LogItem
class LogItem(Item):
#	console_fields = {
#		"Time": "time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(self.ts))",
#		"lvl": "self.lvl",
#		"ts": "self.ts",
#		"line": "self.line",
#	}

	keyfield = "ts"
	def __init__(self, lvl = None, line = None):
		if lvl == None:
			self.ts = None
		else:
			self.ts = time.time()
		self.lvl = lvl
		self.line = line
		super().__init__(self.ts)

	def save(self, withvirtual=False):
		r = {}
		r["ts"] = self.ts
		r["lvl"] = self.lvl
		r["line"] = self.line
		if withvirtual:
			r['Time'] = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(self.ts))
			if len(self.line) > 70:
				r["line"] = self.line[:70] + "..."
		return r

# LogQ
#
class LogQ(Item):
	def __init__(self, conf = None, loop = None):
		self.name = None
		if conf is not None:
			self.name = "logq"
			self.conf = conf
			self.max_log_records = conf.get('max_log_records', 1000)
			self.q = Queue(loop=loop)
			self.loop = loop
		super().__init__(self.name)
		self.prune_in_progress = False
		logger.info("%s logq init", self)

	def save(self, withvirtual=False):
		r = {}
		if withvirtual:
			for t in Table.tables:
				table = Table.tables[t]
				r[ table.tablename + " table"] = len(table)
				try:
					r[ table.tablename + " cache"] = table.cache.status()
				except:
					pass
		return r

	def write(self, msg):
		asyncio.ensure_future(self.awrite(msg), loop=self.loop)

	async def awrite(self, msg):
		msg = msg.rstrip('\n')
		if len(msg) > 0:
			await self.q.put(msg)

	def flush(self):
		pass


	async def asyncfeed(self, ilist):
		for i in ilist:
			yield i

	async def delitem(self, item):
		item.delete()

	async def prune_logitem_table(self, count):
		logger.debug("starting prune of %s item", count)
		csk = CSearchKey(
				table_name=LogItem._table.tablename,
				key_field="ts",
				start_key=None, 
				start_item_number=0, 
				count=count)
		for item in LogItem._table.get_range(csk):
			item.delete()
		logger.debug("prune done, table size %s", len(LogItem._table))
		self.prune_in_progress = False


	async def start(self):
		LogItem._table = DbTable(LogItem, keyfield="ts", tablename="LogItem")
		logger.info("%s logq start", self)
		ansi_escape = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')    # removes ansi escape sequence
		while True:
			msg = await self.q.get()
			if msg is None:
				break
			if len(msg) == 0:
				continue

			try:
				lvl, line = msg.split(None, 1)
				who, what = line.split(None,1)
				if who in ['asyncio_server','asyncio_socket','server','web_protocol']:	# don't log these
					continue
			except:
				continue		# loose badly formated log lines
			line = ansi_escape.sub('', line)

			logitem = LogItem(lvl, line)
			if lvl in ['INFO', 'WARNING', 'ERROR', 'CRITICAL', 'ALERT', 'EMERGENCY']:
				_WEBAPP.send_console_alert(lvl, line)
			count = len(LogItem._table) - self.max_log_records
			if count > 0 and not self.prune_in_progress:
				self.prune_in_progress = True
				asyncio.ensure_future(self.prune_logitem_table(count+10), loop=self.loop)
			self.update()
		logger.info("%s logq stop", self)




#!/usr/bin/env python3

# python library Stealink network

import struct
from asyncio import Queue
from queue import Empty, Full
import socket
import json
import time
import asyncio
import re
import os

import logging
logger = logging.getLogger(__name__)

from .util import phex

from .const import (
    PROJECT_PACKAGE_NAME,
    __version__
)

from .linkage import (
	Item,
	Table,
	DictTable,
	DbTable,
	CSearchKey,
)


MAX_RESEND_COUNT = 25
SL_MAX_MESSAGE_LEN = 255
SL_ACK_WAIT = 3


_MQTT = None
_DB = None
steam_root = None

def Attach(mqtt, db):
	global _MQTT, _DB
	_MQTT = mqtt
	_DB = db
	logger.debug("steamlink: Attached apps '%s, %s'", _MQTT.name, _DB.name)


def set_steam_root(root):
	global steam_root
	steam_root = root

TODO = """
- track routing table from received packets

"""



#
# Exception 
#
class SteamLinkError(Exception):
	def __init__(self, message):

		# Call the base class constructor with the parameters it needs
		super().__init__(message)



SL_RESPONSE_WAIT_SEC = 10
MAX_NODE_LOG_LEN = 1000		# maximum packets stored in per node log



NODEVER = 1
MAXSILENCE = 45

#
# SL_CodeCfgStruct
#
class SL_NodeCfgStruct:
	"""
	Node configuration data, as stored in flash
	defined in SteamLink.h in the steamlink-arduino repo

	struct SL_NodeCfgStruct {
		uint8_t  version;
		uint32_t slid;
		char name[10];
		char description[32];
		float gps_lat;
		float gps_lon;
		short altitude;		// in meters
		uint8_t max_silence; // in seconds
		bool battery_powered;
		uint8_t radio_params; // radio params need to be interpreted by drivers

	"""
	sfmt = '<BL10s32sffhBBB'

	def __init__(self, version = NODEVER, slid = None, name = "*UNK*", description = "*UNK*", gps_lat = 0.0, gps_lon = 0.0, altitude = 0, max_silence = MAXSILENCE, battery_powered = False, radio_params = 0, pkt = None):
		if pkt is None:	 # construct
			self.version = version
			self.slid = slid						# L
			self.name = name						# 10s
			self.description = description			# 32s
			self.gps_lat = gps_lat					# f
			self.gps_lon = gps_lon					# f
			self.altitude = altitude				# h
			self.max_silence = max_silence			# B
			self.battery_powered = battery_powered	# B
			self.radio_params = radio_params		# B

		else:			# deconstruct
			if struct.calcsize(SL_NodeCfgStruct.sfmt) != len(pkt):
				logger.error("NodeCfgStruct: packed messages length incorrect, wanted %s, got %s", struct.calcsize(SL_NodeCfgStruct.sfmt), len(pkt))
				raise SteamLinkError("packed messages length incorrect")
			self.version, self.slid, name, description, self.gps_lat, self.gps_lon, self.altitude, self.max_silence, battery_powered, self.radio_params = struct.unpack(SL_NodeCfgStruct.sfmt, pkt)
			self.name = name.decode().strip('\0')
			self.description = description.decode().strip('\0')
			self.battery_powered = battery_powered == 1

	def pack(self):
		self.pkt = struct.pack(SL_NodeCfgStruct.sfmt, self.version, self.slid, self.name.encode(), self.description.encode(), self.gps_lat, self.gps_lon, self.altitude, self.max_silence, self.battery_powered, self.radio_params)
		return self.pkt


	def __str__(self):
		try:
			return "NodeCFG: %s %s %s" % (self.slid, self.name, self.description)
		except Exception as e:
			return "NodeCFG: undefined: %s" % e

	def save(self):
		d = {
			'slid': self.slid,
			'name': self.name,
			'description': self.description,
			'gps_lat': self.gps_lat,
			'gps_lon': self.gps_lon,
			'altitude': self.altitude,
			'max_silence': self.max_silence,
			'battery_powered': self.battery_powered,
			'radio_params': self.radio_params
		}
		return d

	def update(self, other):
		if self.slid != other.slid:
			logger.info("Node %s changed slid from %s to %s", self.name, self.slid, other.slid)
			self.slid = other.slid
		if self.name != other.name:
			logger.info("Node %s changed name from %s to %s", self.name, self.name, other.name)
			self.name = other.name
		if self.description != other.description:
			logger.info("Node %s changed description from %s to %s", self.name, self.description, other.description)
			self.description = other.description
		if self.gps_lat != other.gps_lat:
			logger.info("Node %s changed gps_lat from %s to %s", self.name, self.gps_lat, other.gps_lat)
			self.gps_lat = other.gps_lat
		if self.gps_lon != other.gps_lon:
			logger.info("Node %s changed gps_lon from %s to %s", self.name, self.gps_lon, other.gps_lon)
			self.gps_lon = other.gps_lon
		# incomplete

#
# SL_OP op codes
#
class SL_OP:
	'''
	control message types: EVEN, 0 bottom bit
	data message types: ODD, 1 bottom bit
	'''

	DN = 0x30		# data to node, ACK 
	BN = 0x32		# slid precedes payload, bridge forward to node
	GS = 0x34		# get status, reply with SS message
	TD = 0x36		# transmit a test message via radio
	SC = 0x38		# set radio paramter to x, acknowlegde with AK
	BC = 0x3A		# restart node, no reply
	BR = 0x3C		# reset the radio, TBD
	AN = 0x3E		# Ack from store -> node

	DS = 0x31		# data to store
	BS = 0x33		# bridge to store
	ON = 0x35		# go online
	AS = 0x37		# acknowlegde the last control message
	OF = 0x39		# go offline
	TR = 0x3B		# Received Test Data
	SS = 0x3D		# status info and counters
	NC = 0x3F		# No Connection or timeout

	def code(code):
		try:
			return list(SL_OP.__dict__.keys())[list(SL_OP.__dict__.values()).index(code)]
		except:
			pass
		return '??'

	def val(val):
		try:
			return SL_OP.__dict__[val]
		except:
			pass
		return 0x99	

SL_AS_CODE = {0: 'Success', 1: 'Supressed duplicate pkt', 2: 'Unexpected pkt, dropping'}

class WaitForAck:

	def __init__(self, waittime):
		self.waittime = waittime
		self.clear_wait()


	def __str__(self):
		if self.waituntil == 0:
			return "NoWait"
		return "Wait %s %s" % (int(self.waituntil), self.pkt)


	def clear_wait(self):
		self.waituntil = 0
		self.pkt = None
		self.count = 0
		self.do_insert = False


	def stop_wait(self):
		""" N.B. returns original pkt if do_insert was set, for db insert """
		pkt = None
		if self.pkt is None:
			if 'waitack' in logging.DBGK: logger.info("wait: redundant Ack")
		else:
			if 'waitack' in logging.DBGK: logger.debug("wait on %s got Ack", self)
			if self.do_insert:
				pkt = self.pkt
		self.clear_wait()
		return pkt


	def restart_wait(self):
		self.waituntil = time.time() + self.waittime
		return self.pkt		# for resend


	def wait_remaining(self):
		if self.waituntil == 0:
			return 0
		return self.waituntil - time.time()


	def set_wait(self, pkt, do_insert=False):
		self.waituntil = time.time() + self.waittime
		self.pkt = pkt
		self.count = 0
		self.do_insert = do_insert
		if 'waitack' in logging.DBGK: logger.debug("wait on %s for %s sec", self.pkt, self.waittime)


	def inc_resend_count(self):
		self.count += 1
		return self.count


	def is_waiting(self):
		return self.waituntil


#
# Steam
#
class Steam(Item):
	def __init__(self, conf = None):
		self.steam_id = 0
		self.autocreate = False
		if conf:
			self.desc = conf['description']
			self.autocreate = conf['autocreate']
			self.name = conf['name']
			self.steam_id = int(conf['id'])
		super().__init__(self.steam_id)
		self.cpubusy = 0
		self.my_ip_address = socket.gethostbyname(socket.gethostname())
		self.identity = "%s %s (%s)" % (PROJECT_PACKAGE_NAME, __version__, self.my_ip_address)
		self._public_topic_control_re = None

		_MQTT.set_msg_callback(self.on_data_msg)

		self._public_topic_control = _MQTT.get_public_control_topic()
		if self._public_topic_control is not None:
			_MQTT.set_public_control_callback(self.on_public_control_msg)
			self._public_topic_control_re = self._public_topic_control % "(.*)"

		self._mqtt_test_succeeded  = False

		mq_cmd_msg = { "cmd": "selfcheck", "identity": self.identity }
		_MQTT.publish("store", json.dumps(mq_cmd_msg), sub="data")


	def set_loglevel(self, loglevel):
		from .linkage import logger as linkage_logger 
		from .mqtt import logger as mqtt_logger 
		from .db import logger as db_logger 
		from .testdata import logger as testdata_logger 
		from .web import logger as web_logger

		logger.setLevel(loglevel)
		linkage_logger.setLevel(loglevel)
		mqtt_logger.setLevel(loglevel)
		db_logger.setLevel(loglevel)
		testdata_logger.setLevel(loglevel)
		web_logger.setLevel(loglevel)


	def handle_web_command(self, sid, message):
		# dict with: cmd, slid, data
		# cmds are: DN SC
		if message['cmd'] in ["DN", "SC"]:
			node = Node._table.find_one(message["slid"])
			if node is None:
				return {"Success": False, "Error": "Unknown node" }
			to_send = message['data']
			if message['cmd'] == "DN":
				ret = node.send_data_to_node(to_send+'\0') 	

			elif message['cmd'] == "SC":
				try:
					nodecfg = json.loads(to_send)
				except Exception as e:
					return {"Success": False, "Error": "nodecfg problem:  %s" % e }
				node.nodecfg = SL_NodeCfgStruct(node.slid, **nodecfg)
				ret = node.send_set_config() 

		if ret is not "OK":
			return {"Success": False,
					"Error": "packet %s was not send to %s: %s" % (to_send, node.name, ret) }
		return {"Success": True }


	def handle_store_command(self, cmd):
		if type(cmd) != type({}):
			logger.warning("unreadable cmd %s", cmd)
			return
		if cmd['cmd'] == 'selfcheck':
			id = cmd.get('identity')
			if id != self.identity:
				logger.warning("there is another system: %s", id)
			else:
				logger.debug("mqtt test successfull")
				self._mqtt_test_succeeded = True
		elif cmd['cmd'] == 'gelog':
			count = cmd.get('count', 1)
			for i in range(count):
				logger.info("useless log entry, just to take up space, num %s", i)
		elif cmd['cmd'] == 'ping':
			response = "pong '%s'" % self.identity
			_MQTT.publish("store", response, sub="control")
		elif cmd['cmd'] == 'debug':
			dbglvl = cmd.get('dbglvl', None)
			slvl = cmd.get('level', None)
			if slvl is not None:
				loglevel = getattr(logging, slvl.upper())
				self.set_loglevel(loglevel)
				logger.warning("setting loglevel to %s", loglevel)
			if dbglvl is not None:
				logging.DBG = int(dbglvl)
		elif cmd['cmd'] == 'shutdown':
			from .main import GracefulExit, GracefulRestart
			if cmd.get('restart',False):
				raise GracefulRestart
			else:
				raise GracefulExit
		else:
			logger.warning("unknown store command %s", cmd)
				

	def on_public_control_msg(self, client, userdata, msg):
		if self._public_topic_control is None:
			return

		if logging.DBG > 2: logger.debug("on_public_control_msg %s %s", msg.topic, msg.payload)
		match = re.match(self._public_topic_control_re, msg.topic) 
		if match is None:
			logger.warning("topic did not match public control topic: %s %s", topic, self._public_topic_control)
			return
		nodename = match.group(1)
		node = Node._table.find_one(nodename, keyfield = "name")
		if node is None:
			logger.warning("public control: no such node %s: %s", nodename, msg.payload)
			return
		try:
			jmsg = json.loads(msg.payload.decode('utf-8'))
			to_send = jmsg['payload'].encode()
		except Exception as e:
			to_send = str(msg.payload.decode('utf-8'))
#			raise
		ret = node.send_data_to_node(to_send+'\0') 	# XXX should not need the b'\0' but nodes crash w/o
		if ret is not "OK":
			logger.warning("packet %s was not send to %s: %s", to_send, node.name, ret)


	def on_data_msg(self, client, userdata, msg):
		# msg has  topic, payload, retain

		topic_parts = msg.topic.split('/', 2)
		if topic_parts[1] == "store":
			try:
				cmd = json.loads(msg.payload.decode('utf-8'))
			except:
				cmd = msg.payload
			logger.debug("store command %s", cmd)
			self.handle_store_command(cmd)
			return

		if logging.DBG > 2: logger.debug("on_data_msg  %s %s", msg.topic, msg.payload)
		try:
			sl_pkt = Packet(pkt=msg.payload)
		except SteamLinkError as e:
			logger.warning("mqtt: pkt dropped: '%s', steamlink error %s", msg.payload, e)
			return
		except ValueError as e:
			logger.warning("mqtt: pkt dropped: '%s', value error %s", msg.payload, e)
			return
		if sl_pkt.slid is None:
			logger.warning("mqtt: pkt dropped: '%s', no slid %s", msg.payload, sl_pkt)
			return

		node = Node._table.find_one(sl_pkt.slid)
		if node is None:		# Auto-create node
			if self.autocreate and sl_pkt.sl_op == SL_OP.ON:
				node = Node(sl_pkt.slid, sl_pkt.nodecfg)
			else:
				logger.warning("on_data_msg: no node for pkt %s", sl_pkt)
				return
		sl_pkt.set_node(node)
		node.post_data(sl_pkt)


	async def start(self):
		process_time = time.process_time()
		now = time.time()
		delta = 0
		wait = 1
		logger.info("%s starting heartbeat", self.name)
		FLUSHWAIT = 10
		flushwait = FLUSHWAIT
		while True:
			await asyncio.sleep(wait)
			self.heartbeat()
			flushwait -= 1
			if flushwait == 0:
				_DB.flush()		# N.B. expensive
				flushwait = FLUSHWAIT

			n_process_time = time.process_time()
			n_now = time.time()

			delta = n_now - now
			wait = 1 - (n_now % 1)
			self.cpubusy = ((n_process_time - process_time) / delta ) * 100.0
			now = n_now
			process_time = n_process_time
#			if logging.DBG == 0:	# N.B. reduce noise when debuging, i.e. no heartbeat
			self.update(True)


	def heartbeat(self):
		for node in Node._table.find(1, 'mesh_id'):
			node.periodic_check()


	def save(self, withvirtual=False):
		r = {}
		r['steam_id'] = self.steam_id
		r['name'] = self.name
		r['desc'] = self.desc
		if withvirtual:
			r['Name'] = self.name
			r['Description'] = self.desc
			r['Meshes'] = len(Mesh._table)
			r['Time'] = time.asctime()
			r['Load'] = "%3.1f%%" % self.cpubusy
			r['Mesh records'] = len(Mesh._table)
			r['Node records'] = len(Node._table)
			r['Packet records'] = len(Packet._table)
		return r

#
# Mesh
#
class Mesh(Item):
	keyfield = 'mesh_id'

	def __init__(self, mesh_id=None):
		self.mesh_id = mesh_id
		if mesh_id is None:
			self.name = "Mesh??" 
		else:
			self.name = "Mesh%s" % self.mesh_id
		self.desc = "Description for %s" % self.name
		self.steam_id = 0	# XXX
		self.packets_sent = 0
		self.packets_received = 0
		super().__init__(self.mesh_id)
		if self.mesh_id is not None:
			logger.info("Mesh created: %s", self)


	def save(self, withvirtual=False):
		r = {}
		r['mesh_id'] = self.mesh_id
		r['steam_id'] = self.steam_id
		r['name'] = self.name
		r['desc'] = self.desc
		if withvirtual:
			r['Name'] = self.name
			r['Description'] = self.desc
			r["Total Nodes"] = len(Node._table)
			r["Active Nodes"] = len(Node._table)
			r["Packets sent"] = self.packets_sent
			r["Packets received"] = self.packets_received
		return r

#
# Node
#
class Node(Item):
	keyfield = 'slid'
	UPSTATES = ["ONLINE", "OK", "UP", "TRANSMITTING"]
	OPS_need_ack = [SL_OP.DS, SL_OP.ON]

	""" a node in a mesh set """
	def __init__(self, slid=None, nodecfg = None):
		if slid is not None:
			slid = int(slid)
			logger.debug("Node creating : %s" % slid)
			if nodecfg is None:
				self.nodecfg = SL_NodeCfgStruct(slid, "Node%08x" % slid)
				logger.debug("Node config is %s", self.nodecfg)
			elif type(nodecfg) == type({}):
				self.nodecfg = SL_NodeCfgStruct(slid, **nodecfg)
				logger.debug("Node config is %s", self.nodecfg)
			else:
				self.nodecfg = nodecfg
			self.name = nodecfg.name
			self.slid = slid
			self.mesh_id = (slid >> 8)
			self.set_mesh()
		else:
			self.name = ""
			self.slid = None
			self.mesh_id = None
		self._response_q = Queue(maxsize=1)

		self.packets_sent = 0
		self.packets_received = 0
		self.packets_resent = 0
		self.packets_dropped = 0
		self.packets_missed = 0
		self.packets_duplicate = 0
		self.pkt_numbers = {True: 0, False:  0}	# next pkt num for data, control pkts
		self.state = "INITIAL"
		self.last_node_restart_ts = 0
		self.last_packet_rx_ts = 0
		self.last_packet_tx_ts = 0
		self.last_packet_num = 0
		self.last_control_pkt = None
		self.last_data_pkt = None
		self.via = []		# not initiatized
		self.tr = {}		# dict of sending nodes, each holds a list of (pktno, rssi)
		self.wait_for_AS = WaitForAck(SL_ACK_WAIT)
#		self.packet_log = TimeLog(MAX_NODE_LOG_LEN)

		super().__init__(slid)
		if slid is not None:
			logger.info("Node created: %s" % self)


	def set_mesh(self):
		self.mesh = Mesh._table.find_one(self.mesh_id)
		if self.mesh is None:		# Auto-create Mesh
			logger.debug("Node %s: mesh %s autocreated", self.slid, self.mesh_id)
			self.mesh = Mesh(self.mesh_id)


	def set_pkt_number(self, pkt):
		dc =  pkt.is_data()
		self.pkt_numbers[dc] += 1
		if self.pkt_numbers[dc] == 0:
			self.pkt_numbers[dc] += 1	# skip 0
		return self.pkt_numbers[dc]


	def load(self, data):	#N.B.
		if 'nodecfg' in data:
			data['nodecfg'] = SL_NodeCfgStruct(**data['nodecfg'])
		super().load(data)
		self.set_mesh()


	def save(self, withvirtual=False):
		r = {}
		r['name'] = self.name
		r[Node.keyfield] = self.slid
		r[Mesh.keyfield] = self.mesh_id
		r['via'] = self.via
		r['nodecfg'] = self.nodecfg.save()
		if withvirtual:
			r['State'] = self.state
			r['Name'] = self.nodecfg.name
			r['Description'] = self.nodecfg.description
			r['Last Pkt received'] = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(int(self.last_packet_rx_ts)))
			r['Last Pkt sent'] = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(int(self.last_packet_tx_ts)))
			r['Last Node restart'] = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(int(self.last_node_restart_ts)))
			r['Packets sent'] = self.packets_sent
			r['Packets received'] = self.packets_received
			r['Packets resent'] = self.packets_resent
			r['Packets dropped'] = self.packets_dropped
			r['Packets missed'] = self.packets_missed
			r['Packets duplicate'] = self.packets_duplicate
			r['wait_for_AS'] = str(self.wait_for_AS)
			r['gps_lat'] = self.nodecfg.gps_lat
			r['gps_lon'] = self.nodecfg.gps_lon
			if self.last_data_pkt is None:
				r['last_data_pkt'] = {}
			else:
				r['last_data_pkt'] = self.last_data_pkt.save(True)
			if self.last_control_pkt is None:
				r['last_control_pkt'] = {}
			else:
				r['last_control_pkt'] = self.last_control_pkt.save(True)
		return r


	def get_firsthop(self):
		if len(self.via) == 0:
			firsthop = self.slid
		else:
			firsthop = self.via[0]
		return firsthop


	def set_state(self, new_state):
		was_up = self.is_state_up()
		old_state = self.state
		self.state = new_state
		is_state_up = self.is_state_up()

		if not was_up and is_state_up:
			logger.info("node %s now online: %s -> %s", self, old_state, new_state)
		elif was_up and not is_state_up:
			logger.info("node %s now offline: %s -> %s", self, old_state, new_state)
		if was_up != is_state_up:
			# publish node state on some mqtt
			pass

#		if new_state == "TRANSMITTING":		#XXX check if node is sleeping or offline
#			self.send_get_status()
		self.update()


	def is_offline(self):
		return self.state == "OFFLINE"


	def is_state_up(self):
		return self.state in Node.UPSTATES


	def is_overdue(self):
		if self.state == "OFFLINE":
			return False
		return (self.last_packet_rx_ts + self.nodecfg.max_silence) <= time.time()


	def publish_pkt(self, sl_pkt=None, resend=False, sub="control"):
		if resend:
			logger.debug("resending pkt: %s", sl_pkt)
			self.packets_resent += 1
		else:
			if self.wait_for_AS.is_waiting() and sl_pkt.sl_op != SL_OP.AN:
				logger.error("attempt to send pkt while waiting for AS, ignored: %s", sl_pkt)
				self.packets_dropped += 1
				return
		if len(sl_pkt.pkt) > SL_MAX_MESSAGE_LEN:
			logger.error("publish pkt to long(%s): %s", len(sl_pkt.pkt), sl_pkt)
			return
#		self.log_pkt(sl_pkt)
		self.packets_sent += 1
		self.mesh.packets_sent += 1
		if logging.DBG > 1: logger.debug("publish_pkt %s to node %s", sl_pkt, self.get_firsthop())
		_MQTT.publish(self.get_firsthop(), sl_pkt.pkt, sub=sub)
		self.last_packet_tx_ts = time.time()
		self.update()
		self.mesh.update(True)


	def send_ack_to_node(self, code):
		sl_pkt = Packet(slnode=self, sl_op=SL_OP.AN, payload=chr(code))	
		self.publish_pkt(sl_pkt)
		return


	def send_boot_cold(self):
		sl_pkt = Packet(slnode=self, sl_op=SL_OP.BC)
		self.publish_pkt(sl_pkt)
		return


	def send_get_status(self):
		sl_pkt = Packet(slnode=self, sl_op=SL_OP.GS)
		self.publish_pkt(sl_pkt)
		return


	def send_data_to_node(self, data): 
		if not self.is_state_up():
			self.packets_dropped += 1
			return "NodeDown"

		if self.wait_for_AS.is_waiting():
			self.packets_dropped += 1
			return "AckWait"

		bpayload = data
		logger.debug("send_data_to_node:: len %s, pkt %s", len(bpayload), bpayload)
		sl_pkt = Packet(slnode=self, sl_op=SL_OP.DN, payload=bpayload)
		self.last_control_pkt = sl_pkt
		self.publish_pkt(sl_pkt)
		self.wait_for_AS.set_wait(sl_pkt, do_insert=True)
		sl_pkt.insert()
		return "OK"


	def send_set_config(self): 
		if not self.is_state_up():
			self.packets_dropped += 1
			return "NodeDown"

		if self.wait_for_AS.is_waiting():
			self.packets_dropped += 1
			return "AckWait"

		bpayload = self.nodecfg.pack()
		logger.debug("send_set_config: len %s, pkt %s", len(bpayload), self.nodecfg)
		sl_pkt = Packet(slnode=self, sl_op=SL_OP.SC, payload=bpayload)
		self.publish_pkt(sl_pkt)
		self.wait_for_AS.set_wait(sl_pkt)
		return "OK"


	def send_testpacket(self, pkt):
		if not self.is_state_up():
			self.packets_dropped += 1
			return SL_OP.NC
		sl_pkt = Packet(slnode=self, sl_op=SL_OP.TD, payload=pkt)
		self.publish_pkt(sl_pkt)
		rc = self.get_response(timeout=SL_RESPONSE_WAIT_SEC) # No!
		logger.debug("send_packet %s got %s", sl_pkt, SL_OP.code(rc))
		return rc


	def check_duplicate_pkt_num(self, sl_pkt):
		pkt_num = sl_pkt.pkt_num
		last_packet_num = self.last_packet_num
		self.last_packet_num = pkt_num
		if last_packet_num == 0:			# we did not see packets from this node before
			return False
		if pkt_num == 0xFFFF:				# wrap
			self.last_packet_num = 0		# remote will skip 0
		if pkt_num == last_packet_num + 1:	# proper sequence
			return False
		if pkt_num == last_packet_num:		# duplicate
			logger.info("%s: received duplicate pkt %s", self, sl_pkt)
			return True
		if pkt_num != 1:					# remote restarted
			missed = pkt_num-(last_packet_num+1)
			self.packets_missed += missed
			logger.error("%s: %s pkts missed before %s", self, missed, sl_pkt)
		return False


	def store_data(self, sl_pkt):
		if sl_pkt.sl_op != SL_OP.DS:		# actual data
			logger.warning("store_data NOT storing non-DS data: %s", sl_pkt.sl_op)
			return
		if logging.DBG >= 1: logger.debug("store_data inserting into db")

		sl_pkt.insert()
		self.send_ack_to_node(0)

		_MQTT.public_publish(self.name, sl_pkt.payload)
		

	def post_data(self, sl_pkt):
		""" handle incoming messages on the ../data topic """
		if sl_pkt.is_data():
			if self.check_duplicate_pkt_num(sl_pkt):	# duplicate packet
				if sl_pkt.sl_op in Node.OPS_need_ack:
					logger.debug("post_data send AN on duplicate %s", SL_OP.code(sl_pkt.sl_op))
					self.send_ack_to_node(0)
				self.packets_duplicate += 1
				self.packets_dropped += 1
				return	# duplicate
		else:
			logger.error("%s got control pkt %s", self, sl_pkt)
			self.packets_dropped += 1
			return # NotForUs

		# set ts for all nodes on the via route
		for slid in sl_pkt.via:
			node = Node._table.find_one(slid)
			if node:
				node.last_packet_rx_ts = sl_pkt.ts
				if not node.is_state_up():
					node.set_state('TRANSMITTING')
				node.update()
			else:
				self.packets_dropped += 1
				logger.error("post_data: via node %s not on file", slid)

		# check for routing changes
		if self.via == []:
			self.via = sl_pkt.via
		elif self.via != sl_pkt.via:
			logger.warning("node %s routing changed, was %s is now %s", \
					self, self.via, sl_pkt.via)
			self.via = sl_pkt.via

		self.packets_received += 1
		self.mesh.packets_received += 1

		# logger.info("%s: received %s", self, sl_pkt)

		self.tr[sl_pkt.slid] = sl_pkt.rssi

		sl_op = sl_pkt.sl_op

		if sl_op == SL_OP.ON: # autocreate did set nodecfg
			self.wait_for_AS.clear_wait()		# give up 
			logger.debug('post_data: slid %d ONLINE', int(self.slid))

			new_nodecfg = SL_NodeCfgStruct(pkt=sl_pkt.bpayload)
			self.nodecfg.update(new_nodecfg)			
			self.send_set_config()
			self.set_state("ONLINE")
			self.last_node_restart_ts = time.time()
			logger.info('%s signed on', self)
		elif sl_op == SL_OP.OF:
			self.set_state("OFFLINE")

		elif sl_op == SL_OP.DS:
#			logger.debug('post_data: slid %d status %s', int(self.slid),sl_pkt.payload)
			self.last_data_pkt = sl_pkt
			self.store_data(sl_pkt)

		elif sl_op == SL_OP.SS:
#			logger.debug("post_data: slid %d status '%s'", int(self.slid),sl_pkt.payload)
			self.set_state(sl_pkt.payload)

		elif sl_op == SL_OP.AS:
			logger.debug('post_data: slid %d ACK:  %s', int(self.slid), 
						SL_AS_CODE[int(sl_pkt.bpayload[0])])
			pkt = self.wait_for_AS.stop_wait()		# done waiting, discard stashed packet
			if pkt is not None:
				pkt.insert()					# store pkt

		elif sl_op == SL_OP.TR:
			logger.debug('post_data: node %s test msg', sl_pkt.payload)

			try:
				test_pkt = TestPkt(pkt=sl_pkt.payload)
			except ValueError as e:
				logger.warning("post_incoming: cannot identify test data in %s", sl_pkt.payload)
				self.packets_dropped += 1
				return

			test_pkt.set_receiver_slid(sl_pkt.via)
			test_pkt.set_rssi(sl_pkt.rssi)
			if not test_pkt.pkt['slid'] in self.tr:
				self.tr[test_pkt.pkt['slid']] = []
			self.tr[test_pkt.pkt['slid']].append((test_pkt.pkt['pktno'], test_pkt.pkt['rssi']))
#			sl_log.post_incoming(test_pkt)

		self.last_packet_rx_ts = sl_pkt.ts

		# any pkt from node indicates it's up
		if not self.is_offline and not self.is_state_up():
			self.set_state('TRANSMITTING')

		self.update()
		self.mesh.update(True)


	def periodic_check_for_AS(self):
		if self.wait_for_AS.is_waiting():
			rwait = self.wait_for_AS.wait_remaining()
			logger.debug("periodic_check: %s wait %s sec for AS ", self.name, rwait)
			if rwait <= 0:
				if self.wait_for_AS.inc_resend_count() > MAX_RESEND_COUNT:
					logger.info("resend limit reached for %s, giving up", self.wait_for_AS)
					self.wait_for_AS.clear_wait()
				else:
					pkt = self.wait_for_AS.restart_wait()
					self.publish_pkt(pkt, resend=True)


	def periodic_check(self):
		self.periodic_check_for_AS()
		if self.is_overdue() and self.is_state_up():
			self.set_state("OVERDUE")
		if not self.is_offline() and not self.is_state_up():	#XXX not offline or sleeping
			if self.last_packet_tx_ts != 0 and self.last_packet_tx_ts + MAXSILENCE < time.time():
				self.send_get_status()

	def get_response(self, timeout):
		try:
			data = self._response_q.get(timeout=timeout)
		except Empty:
			data = SL_OP.NC
		return data



#
# Packet
#
class BasePacket:
	keyfield = 'ts'
	data_header_fmt = '<BLHB%is'		# op, slid, pkt_num, rssi, payload"
	control_header_fmt = '<BLH%is'		# op, slid, pkt_num, payload"

	def __init__(self, slnode, sl_op = None, rssi = 0, payload = None, pkt = None):

		self.slid = None
		self.rssi = 0
		self.via = []
		self.pkt_num = None
		self.payload = None
		self.ts = time.time()
		self.nodecfg = None
		if pkt is None and sl_op is None:
			return	# needs a load() to complete
		self.is_outgoing = pkt is None

		if self.is_outgoing:				# construct pkt
			self.slid = slnode.slid
			self.construct(slnode, sl_op, rssi, payload)
		else:								# deconstruct pkt
			if not self.deconstruct(pkt):
				logger.error("deconstruct pkt to short: %s", len(pkt))
				raise SteamLinkError("deconstruct pkt to short");


	def __str__(self):
		ULOn = "[4m"
		BOn = "[7m"
		BOff = "[0m"
		try:
			return "Packet N%s(%s)%s" % ( self.slid, self.pkt_num, BOn+SL_OP.code(self.sl_op)+BOff)
		except Exception as e:
			return "Packet NXXX(%s)??" % e


	def is_data(self, sl_op = None):
		if sl_op is None:
			sl_op = self.sl_op
		if sl_op is None:
			logger.error("packet op not yet known")
			raise SteamLinkError("packet op not yet known");
		return (sl_op & 0x1) == 1


	def construct(self, slnode, sl_op, rssi, payload):
		self.slid = slnode.slid
		self.sl_op = sl_op
		self.rssi = rssi + 256
		if self.sl_op == SL_OP.ON:
			self.payload = payload.pack()	# payload is a nodecfg 
		else:
			self.payload = payload
		if logging.DBG > 2: logger.debug("SteamLinkPacket payload = %s", payload);
		if self.payload is not None:
			if type(self.payload) == type(b''):
				self.bpayload = self.payload
			else:
				self.bpayload = self.payload.encode('utf8')
		else:
			self.bpayload = b''

		self.pkt_num = slnode.set_pkt_number(self)
		if self.is_data():	# N.B. store never sends data
			sfmt = Packet.data_header_fmt % len(self.bpayload)
			if logging.DBG > 0: logger.debug("pack: %s %s %s %s %s", SL_OP.code(self.sl_op), \
					self.slid, self.pkt_num, self.rssi, self.bpayload)
			self.pkt = struct.pack(sfmt,
					self.sl_op, self.slid, self.pkt_num, 256 - self.rssi, self.bpayload)
		else:
			sfmt = Packet.control_header_fmt % len(self.bpayload)
			self.pkt = struct.pack(sfmt,
					self.sl_op, self.slid, self.pkt_num, self.bpayload)
			if len(slnode.via) > 0:
				for via in slnode.via[::-1]:
					self.bpayload = self.pkt
					sfmt = Packet.control_header_fmt % len(self.bpayload)
					self.pkt = struct.pack(sfmt, SL_OP.BN, via, 0, self.bpayload)
			if logging.DBG > 1:
				for l in phex(self.pkt, 4):
					logger.debug("pkt c:  %s", l)



	def deconstruct(self, pkt):
		self.pkt = pkt
		if logging.DBG > 1:
			for l in phex(pkt, 4):
				logger.debug("pkt:  %s", l)

		if pkt[0] == SL_OP.BS:		# un-ecap all
			while pkt[0] == SL_OP.BS:
				payload_len = len(pkt) - struct.calcsize(Packet.data_header_fmt % 0)
				sfmt = Packet.data_header_fmt % payload_len
				self.sl_op, slid, self.pkt_num, self.rssi, self.bpayload \
						= struct.unpack(sfmt, pkt)

				self.via.append(slid)
				pkt = self.bpayload
				if logging.DBG > 1: logger.debug("pkt un-ecap BS from P%s(%s)BS, len %s rssi %s", slid, self.pkt_num,  len(pkt), 256-self.rssi)
			self.rssi = self.rssi - 256

		if len(pkt) < struct.calcsize(Packet.data_header_fmt % 0):
			logger.error("deconstruct pkt to short: %s", len(pkt))
			return False;
		if self.is_data(pkt[0]):
			payload_len = len(pkt) - struct.calcsize(Packet.data_header_fmt % 0)
			sfmt = Packet.data_header_fmt % payload_len
			self.sl_op, self.slid, self.pkt_num, rssi, self.bpayload \
						= struct.unpack(sfmt, pkt)
		else:
			payload_len = len(pkt) - struct.calcsize(Packet.control_header_fmt % 0)
			sfmt = Packet.control_header_fmt % payload_len
			self.sl_op, self.slid, self.pkt_num, self.bpayload \
						= struct.unpack(sfmt, pkt)
		self.payload = None

		if self.sl_op == SL_OP.ON:
			try:
				self.nodecfg = SL_NodeCfgStruct(pkt=self.bpayload)
			except SteamLinkError as e:
				logger.error("deconstruct: %s", e)
				return False
			logger.debug("Node config is %s", self.nodecfg)

		if len(self.bpayload) > 0:
			try:
				self.payload = self.bpayload.decode('utf8').strip('\0')
			except Exception as e:
				pass

		return True


#
# Packet
#
class Packet(BasePacket, Item):
	def __init__(self, slnode = None, sl_op = None, rssi = 0, payload = None, pkt = None, _load=None):

		BasePacket.__init__(self, slnode, sl_op, rssi, payload, pkt)
		if pkt is None and slnode is None:
			return	# needs a load() to complete
		self.is_outgoing = pkt is None

		if self.is_outgoing:				# construct pkt
			self.set_node(slnode)
		(Item).__init__(self, None, _load)


	def set_node(self, node):
		if logging.console:
			ULOn = "[4m"
			BOn = "[7m"
			BOff = "[0m"
		else:
			ULOn = ""
			BOn = ""
			BOff = ""

		self.node = node
		if self.is_outgoing:
			direction = "send"
			via = "direct" if self.node.via == [] else "via %s" % self.node.via
		else:
			direction = "received"
			via = "direct" if self.via == [] else "via %s" % self.via

		logger.debug("pkt: %s %s %s: %s", ULOn+ direction, via+BOff,  self, self.payload)


	def load(self, data):
		super().load(data)
		self.sl_op = SL_OP.val(self.sl_op)	# xlate from 2-letter-code to val


	def save(self, withvirtual=False):
		r = {}
		r['sl_op'] = SL_OP.code(self.sl_op)
		r['pkt_num'] = self.pkt_num
		r['slid'] = self.slid
		r[Packet.keyfield] = self.ts
		r['rssi'] = self.rssi
		r['ts'] = self.ts
		r['via'] = self.via
		if type(self.payload) == type(b''):
			r['payload'] = repr(self.payload)
		else:
			r['payload'] = self.payload
		if withvirtual:
			r["Time"] = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(self.ts))
			r["op"] = SL_OP.code(self.sl_op)
		return r


#
# TestPkt
#
class TestPkt:
	packet_counter = 1
	def __init__(self, gps=None, text=None, from_slid=None, pkt=None):
		self.pkt = {}
		if text != None and from_slid != None:	# construct pkt
#			self.pkt['lat'] = gps['lat']
#			self.pkt['lon'] = gps['lon']
			self.pkt['slid'] = from_slid
			self.pkt['pktno'] = TestPkt.packet_counter
			self.pkt['text'] = text
			self.pkt['directon'] = 'send'
			TestPkt.packet_counter += 1
		else:									# deconstruct string
			r = pkt.split('|',4)
			self.pkt['lat'] = float(r[0])
			self.pkt['lon'] = float(r[1])
			self.pkt['slid'] = int(r[2])
			self.pkt['pktno'] = int(r[3])
			self.pkt['directon'] = 'recv'
			self.pkt['text'] = r[4]
		ts = time.strftime("%Y-%m-%d_%H:%M:%S", time.localtime())
		self.pkt[Packet.keyfield] = ts


	def get_pktno(self):
		return self.pkt['pktno']


	def set_receiver_slid(self, recslid):
		self.pkt['recslid'] = recslid


	def set_rssi(self, rssi):
		self.pkt['rssi'] = rssi


	def pkt_string(self):
		return "%(lat)0.4f|%(lon)0.4f|%(slid)s|%(pktno)s|%(text)s" % self.pkt


	def json(self):
		return json.dumps(self.pkt)


	def __str__(self):
		return "TESTP(%s)" % str(self.pkt)


#
# NodeRoutes
#
class NodeRoutes:
	def __init__(self, dest, via):
		self.dest = dest
		self.via = via


	def __str__(self):
		svia = ""
		for v in self.via:
			svia += "->0x%02x" % v
		return "VIA(%d: %s" % (self.dest, svia)


#
# LogData
#
class LogData:
	""" Handle incoming pkts on the ../data topic """
	def __init__(self, conf):
		self.conf = conf
		self.logfile = open(conf["file"],"a+")
		self.pkt_inq = Queue()
		self.nodes_online = 0


	def log_state(self, slid, new_state):
		logger.debug("logdata node %d %s", slid, new_state)
		self.nodes_online += 1 if new_state == "ONLINE" else -1


	def post_incoming(self, pkt):
		""" a pkt arrives """

		self.log_pkt(pkt, "receive")
		self.pkt_inq.put(pkt, "recv")


	def post_outgoing(self, pkt):
		""" a pkt is sent """
		self.log_pkt(pkt, "send")


	def log_pkt(self, pkt, direction):
		self.logfile.write(pkt.json()+"\n")
		self.logfile.flush()


	def wait_pkt_number(self, pktnumber, timeout, num_packets):
		""" wait for pkt with number pktnumber for a max of timeout seconds """
		lwait = timeout
		packets_seen = 0
		while True:
			now = time.time()
			try:
				test_pkt = self.pkt_inq.get(block=True, timeout=lwait)
				packets_seen += 1
			except Empty:
				test_pkt = None
			logger.debug("wait_pkt_number pkt %s", test_pkt)
			waited = time.time() - now
			if test_pkt and test_pkt.pkt['pktno'] == pktnumber and packets_seen == num_packets:
				return pktnumber
			if waited >= lwait or test_pkt.pkt['pktno'] > pktnumber:	# our pkt will never arrive
				return None
			lwait -= waited

"""
 message = {
     'table_name':
	 'key_field':
	 'restrict_by': 
	 'start_key':
	 'start_item_number':
	 'end_key':
	 'count':
	 'stream_tag':
  }
"""


def add_csearch(webnamespace, sid, message):

	table_name = message['table_name']
	try:
		table = Table.tables[table_name]
	except KeyError as e:
		return { 'error': 'Table %s not found. Availabe are %s' % (str(e), list(Table.tables.keys())) }

	csearchkey = CSearchKey(**message)

	try:
		csearchkey = table.add_csearch(webnamespace, csearchkey, sid)
	except KeyError as e:
		msg = 'could not add search %s' % (str(e))
		logger.info('add_search fail: %s', msg)
		return { 'error': msg }

	#  force update
	for cs in table.csearches:
		table.csearches[cs].force_update(sid)

	if logging.DBG > 1: logger.debug("add_csearch sid %s csearchkey %s", sid, csearchkey)
	res = {
			'start_key': csearchkey.start_key,
			'end_key': csearchkey.end_key,
			'count': csearchkey.count,
			'start_item_number': csearchkey.start_item_number,
			'total_item_count':  csearchkey.total_item_count,
			'at_start':  csearchkey.at_start,
			'at_end':  csearchkey.at_end,
		}
	return res


def drop_csearch(webnamespace, sid, message):

	table_name = message.get('table_name', None)
	if table_name == None:		# all all tables, i.e. disconnect
		table_list = list(Table.tables.values())
	else:
		try:
			table = Table.tables[table_name]
		except KeyError as e:
			return { 'error': 'Table %s not found' % str(e) }
		table_list = [table]

	for tab in table_list:
		tab.drop_sid_from_csearch(sid)

	return { 'Success': True }


def run_cmd(webnamespace, sid, message):
	res = steam_root.handle_web_command(sid, message)
	return res


# tables = {}
def SteamSetup():
	Steam._table = DbTable(Steam, keyfield="steam_id", tablename="Steam")
	Mesh._table = DbTable(Mesh, keyfield="mesh_id", tablename="Mesh")
	Node._table = DbTable(Node, keyfield="slid", tablename="Node")
	Packet._table = DbTable(Packet, keyfield="ts", tablename="Packet")





# python library Steamlink 

import asyncio
import os
from collections import  Mapping, OrderedDict

from tinydb import TinyDB, Query, Storage, where
from tinydb.database import  Document
from tinydb.storages import JSONStorage, touch
from tinydb.storages import MemoryStorage
from tinydb.middlewares import CachingMiddleware

import logging
logger = logging.getLogger()

from .linkage import check_restrictions


class DBIndex(OrderedDict):
	def __init__(self, table, keyfield):
		self.keyfield = keyfield
		self.table = table
		super().__init__()


	def init_idx(self):
		for item in self.table:
			key = item[self.keyfield]
			super().__setitem__(key, item)


	def has(self, key):
		return key in self


	def insert(self, item):
		key = item[self.keyfield]
		super().__setitem__(key, item)


	def __del__(self, item):
		key = item[self.keyfield]
		super().__setitem__(key, item)


class DBTable:
	def __init__(self, table, name):
		if logging.DBG > 2: logger.debug("DBTable %s", name)
		self.table = table
		self.name = name
		self.idxs = {}
		self.query = Query()


	def insert(self, rec):
		did = self.table.insert(rec)
		if 'dbops' in logging.DBGK: logger.debug("REC insert %s rec %s, %s: %s", self.name, did, type(rec), rec)


	def upsert(self, rec, keyfield):
		try:
			r = rec[keyfield]
		except:
			logger.error("insert with no '%s' field: %s", keyfield, rec)
			return
		el = self.table.search(where(keyfield) == r)
		if 'dbops' in logging.DBGK: logger.debug("upsert search %s return %s", (where(keyfield) == r), el)
		if el is not None and len(el) > 0:
			if el[0] == rec:
				if 'dbops' in logging.DBGK: logger.debug("upsert --nochange--")
				return
			did = self.table.update(rec, where(keyfield) == r)
			if 'dbops' in logging.DBGK: logger.debug("REC upsert update %s rec %s, %s: %s", self.name, did, type(rec), rec)
		else:
			did = self.table.insert(rec)
			if 'dbops' in logging.DBGK: logger.debug("upsert insert %s rec %s, %s: %s", self.name, did, type(rec), rec)


	def db_update(self, rec, keyfield, key):
		if 'dbops' in logging.DBGK: logger.debug("REC update %s rec %s", self.name, rec)
		did = self.table.update(rec, where(keyfield) == key)

	def delete(self, field, val):
		el = self.table.get(where(field) == val)
		if el is None:
			logger.error("delete in %s, no document with %s=%s", self.name, field, val)
			return
		
		try:
			self.table.remove(eids=[el.eid])
		except KeyError as e:
			logger.error("delete in %s, no docid %s", self.name, el.eid)


	def get(self, field, op, val):
		q = "self.table.get(where('%s') %s %s)" % (field, op, repr(val))
		res = eval(q)
		if 'dbops' in logging.DBGK: logger.debug("get %s rec %s: %s", self.name, q, res)
		return res


	def search(self, field, op, val):
		q = "self.table.search(where('%s') %s %s)" % (field, op, repr(val))
		res = eval(q)
		if 'dbops' in logging.DBGK: logger.debug("search %s rec %s: %s", self.name, q, res)
		return res


	def get_range(self, csk):
		""" get a range of records, obeying restrictions
		- if start_key is null, use start_item_number.
		- if start_item_number is negative start from the end
		if 'get_range' in logging.DBGK: logger.debug("get_range: %s", str(csk))
		"""
		if 'get_range' in logging.DBGK: logger.debug("get_range csk %s", str(csk))
		key_field = csk.key_field
		startv = csk.start_key
		endv = csk.end_key
		count = csk.count

		csk.total_item_count = 0

		if len(self.table) == 0:
			if 'get_range' in logging.DBGK: logger.debug("get_range table empty")
			csk.count = 0
			return {}

		udict = {}
		for t in self.table:		# N.B. ad-hoc index over entrie table: expensive!
			udict[t[key_field]] = t
		fullsdict = sorted(udict)
		if 'get_range' in logging.DBGK: logger.debug("get_range table %s items", len(fullsdict))
	
		if len(csk.restrict_by) == 0:
			sdict = fullsdict
		else:
			sdict = []
			for r in fullsdict:
				if check_restrictions(csk.restrict_by, udict[r]):
					sdict.append(r)
			if 'get_range' in logging.DBGK: logger.debug("get_range restricted %s items", len(sdict))

		if len(sdict) == 0:
			if 'get_range' in logging.DBGK: logger.debug("get_range table empty after destrict")
			return {}

		if startv in [None]:
			if csk.start_item_number < 0:
				sidx = max(0, len(sdict) + csk.start_item_number)
			else:
				sidx = min(csk.start_item_number, len(sdict)-1)
			startv = sdict[sidx]
		else:
			sidx = None
			for idx in range(len(sdict)):
				if sdict[idx] >=  startv:
					sidx = idx
					startv = sdict[sidx]
					break
			if sidx is None:
				if 'get_range' in logging.DBGK: logger.debug("get_range no start key found")
				return {}
		if endv in [None]:
			eidx = min(sidx + count-1, len(sdict)-1)
			endv = sdict[eidx]
		else:
			eidx = None
			for idx in range(sidx, len(sdict)):
				if sdict[idx] <= endv:
					endv = sdict[idx]
					eidx = idx
					break
			if eidx is None:
				if 'get_range' in logging.DBGK: logger.debug("get_range no end key found")
				return {}
			count = eidx - sidx + 1

		csk.start_key = sdict[sidx]
		csk.end_key = sdict[eidx]
		csk.start_item_number = sidx
		csk.total_item_count = len(sdict)
		csk.at_start = csk.start_key == sdict[0]
		csk.at_end = csk.end_key == sdict[-1]

		if 'get_range' in logging.DBGK: logger.debug("get_range size %s", (eidx-sidx+1))
		for idx in range(sidx, eidx+1):
			yield udict[sdict[idx]]


	def __len__(self):
		return len(self.table)



class DB:
	""" Notes:
			- inserts get slow with the standard Json modules, check out ujson
			    ( at 3000 items, the per item insert time is 12ms!! )
			- consider creating a new table/db/fle very day or every x records
			 
	"""
	def __init__(self, conf, loop):
		if logging.DBG > 2: logger.debug("DB %s", conf)
		self.name = "DB"
		self.conf = conf
		self.loop = loop
		self.db = None
		self.db_tables = {}


	async def start(self):

		logger.info("%s opening DB %s", self.name, self.conf['db_filename'])
		self.db = TinyDB(self.conf['db_filename'], \
				sort_keys=True, indent=4, separators=(',', ': '), \
				storage=CachingMiddleware(JSONStorage))


	def table(self, name):
		if name in self.db_tables:
			return self.db_tables[name]
	
		db_table = self.db.table(name)
		table = DBTable(db_table, name)
		self.db_tables[name] = table
		return table


	async def stop(self):
		self.close()


	def close(self):
		logger.info("%s closing DB", self.name)
		for tab in self.db_tables:
			try:
				self.db_tables[tab].close()
				del self.db_tables[tab]
			except:
				pass
		self.db.close()
		self.db = None


	def flush(self):
		self.db._storage.flush()





#!/usr/bin/env python3

# a Python based Node for SteamLink

import asyncio
import time
import sys

import logging
logger = logging.getLogger()

from .mqtt import Mqtt
from .steamlink import (
	SL_OP,
	SL_NodeCfgStruct,
	BasePacket,
	WaitForAck, 
	SL_MAX_MESSAGE_LEN, 
	SL_ACK_WAIT,
	MAX_RESEND_COUNT,
)
from .util import phex



class PyNode:
	def __init__(self, nodecfg, conf_mqtt, loop):
		self.nodecfg = nodecfg
		self.slid = self.nodecfg.slid
		self.conf_mqtt = conf_mqtt
		self.mqtt = None
		self.loop = loop
		logger.info("Start")
		self.receive_q = None
		self.via = []
		self.packets_sent = 0
		self.packets_sent = 0
		self.packets_resent = 0
		self.packets_dropped = 0
		self.packets_missed = 0
		self.packets_duplicate = 0
		self.pkt_num = 0
		self.wait_for_AN = WaitForAck(SL_ACK_WAIT)
		self.wait_handle = None

		self.last_packet_tx_ts = 0

		self.status = "OK"


	async def start(self):
		if self.mqtt is not None:
			logger.error("start called more than once")
			sys.exit(1)
		self.mqtt = Mqtt(self.conf_mqtt, as_node=True)
		logger.debug("startup: create Mqtt")
		self.mqtt.set_msg_callback(self.on_control_msg)
		await self.mqtt.start()
		self.send_online_to_store()
		logger.debug("startup: done")


	def on_control_msg(self, client, userdata, msg):
		if logging.DBG > 0: logger.info("on_control_msg: got %s", msg.payload)
		try:
			pkt = BasePacket(slnode=self, pkt=msg.payload)
		except SteamLinkError as e:
			logger.warning("mqtt: pkt dropped: '%s', steamlink error %s", msg.payload, e)
			return
		except ValueError as e:
			logger.warning("mqtt: pkt dropped: '%s', value error %s", msg.payload, e)
			return

		if pkt.slid != self.slid:
			if logging.DBG > 1: logger.warning("mqtt: pkt '%s', not for us", msg.payload)
			return


		if pkt.sl_op == SL_OP.AN:		# Ack
			if self.wait_handle is not None:
				self.wait_handle.cancel()
				self.wait_handle = None
			if self.wait_for_AN is not None:
				pkt = self.wait_for_AN.stop_wait()

		elif pkt.sl_op == SL_OP.SC:	
			rc = self.handle_sc(pkt.payload)
			
		elif pkt.sl_op == SL_OP.DN:	
			asyncio.ensure_future(self.receive_q.put(pkt.payload), loop=self.loop)
			self.send_ack_to_store(0)

		elif pkt.sl_op == SL_OP.SS:	
			self.handle_sc(pkt.payload)
		else:
			logger.error("mqtt: for now, not handling %s", pkt)


	def process_ack_timeout(self):
		logger.debug("process_ack_timeout")
		self.wait_handle = None
		if self.wait_for_AN.inc_resend_count() > MAX_RESEND_COUNT:
			logger.info("resend limit reached for %s, giving up", self.wait_for_AN)
			self.wait_for_AN.clear_wait()
		else:
			pkt = self.wait_for_AN.restart_wait()
			self.publish_pkt(pkt, resend=True)
	

	def handle_gs(self, payload):
		if logging.DBG > 1: logger.debug("handle gs: %s", payload)
		self.send_status_to_store()


	def handle_sc(self, payload):
		if logging.DBG > 1: logger.debug("handle sc: %s", payload)
		# XXX: actually config  stuff
		self.send_ack_to_store(0)


	def send_data_to_store(self, data):
		sl_pkt = BasePacket(self, sl_op=SL_OP.DS, payload=data)
		self.publish_pkt(sl_pkt)
		self.wait_for_AN.set_wait(sl_pkt)


	def send_ack_to_store(self, code):
		sl_pkt = BasePacket(slnode=self, sl_op=SL_OP.AS, payload=chr(code))
		self.publish_pkt(sl_pkt)


	def send_online_to_store(self):
		sl_pkt = BasePacket(self,  SL_OP.ON, payload=self.nodecfg)
		logger.info("send: %s", sl_pkt)
		if self.publish_pkt(sl_pkt):
			self.wait_for_AN.set_wait(sl_pkt)
			self.wait_handle = self.loop.call_later(SL_ACK_WAIT, self.process_ack_timeout)
			return True
		else:
			return False	


	def send_status_to_store(self):
		sl_pkt = BasePacket(self, sl_op=SL_OP.SS, payload=self.status)
		self.publish_pkt(sl_pkt)


	def set_pkt_number(self, pkt):
		self.pkt_num += 1
		if self.pkt_num == 0:
			self.pkt_num = 1
		return self.pkt_num


	def publish_pkt(self, sl_pkt=None, resend=False, sub="data"):
		if resend:
			if logging.DBG > 1: logger.debug("resending pkt: %s", sl_pkt)
			self.packets_resent += 1
		else:
			if self.wait_for_AN.is_waiting() and sl_pkt.sl_op != SL_OP.AN:
				logger.error("attempt to send pkt while waiting for AN, ignored: %s", sl_pkt)
				self.packets_dropped += 1
				return False
		if len(sl_pkt.pkt) > SL_MAX_MESSAGE_LEN:
			logger.error("publish pkt to long(%s): %s", len(sl_pkt.pkt), sl_pkt)
			return False
		self.packets_sent += 1
		if logging.DBG > 1: logger.debug("publish_pkt %s", sl_pkt )
		self.mqtt.publish(self.slid, sl_pkt.pkt, sub=sub)
		self.last_packet_tx_ts = time.time()
		return True


	# User routins

	def send(self, packet, slid=1):
		logger.info("sending: '%s' to '%s'", packet, slid)
		self.send_data_to_store(packet)


	def register_receive_queue(self, receive_q):
		self.receive_q = receive_q


	def set_status(self, new_status):
		self.status = new_status[:20]
		self.send_status_to_store()





#!/usr/bin/env python3
"""SteamLink setup script."""
import os
from setuptools import setup, find_packages
from steamlink.const import (__version__, PROJECT_PACKAGE_NAME,
								PROJECT_LICENSE, PROJECT_URL,
								PROJECT_EMAIL, PROJECT_DESCRIPTION,
								PROJECT_CLASSIFIERS, GITHUB_URL,
								PROJECT_AUTHOR)

HERE = os.path.abspath(os.path.dirname(__file__))
DOWNLOAD_URL = ('{}/archive/'
				'{}.tar.gz'.format(GITHUB_URL, __version__))

PACKAGES = find_packages(exclude=['tests', 'tests.*'])
REQUIRES = [
	'PyYAML>=3.11,<4',
	'aiohttp>=3.0.0',
	'aiomqtt>=0.1.0',
	'python-socketio>=1.8.4',
	'paho-mqtt>=1.3.1,<2',
	'hbmqtt==0.8.0',
	'jinja2==2.10',
	'aiohttp-jinja2>=1.0.0',
	'tinydb>=3.9.0',
	'tinydb-smartcache>=1.0.2',
	'zeroconf>=0.20.0'
]

setup(
	name=PROJECT_PACKAGE_NAME,
	version=__version__,
	license=PROJECT_LICENSE,
	url=PROJECT_URL,
	download_url=DOWNLOAD_URL,
	author=PROJECT_AUTHOR,
	author_email=PROJECT_EMAIL,
	description=PROJECT_DESCRIPTION,
	packages=PACKAGES,
	include_package_data=True,
	zip_safe=False,
	platforms='any',
	install_requires=REQUIRES,
	keywords=['iot', 'LoRa'],
	entry_points={
		'console_scripts': [
			'steamlink = steamlink.main:steamlink_command'
		]
	},
	classifiers=PROJECT_CLASSIFIERS,
)



#!/usr/bin/env python3

# Main program for a Steamlink network store

import sys
import os
import pathlib
import traceback
import hbmqtt
import hbmqtt.broker
import syslog
import random
import asyncio
import socketio
import signal
import yaml
from collections import  Mapping, OrderedDict

import logging
import logging.handlers
logger = logging.getLogger()

from .linkage import Attach as linkageAttach

# SteamLink project imports
from .mqtt import (
	Mqtt,
	Mqtt_Broker
)

from .linkage import LogQ
from .linkage import DictTable
from .steamlink import SteamSetup, Steam, Mesh, Node, set_steam_root
from .steamlink import Attach as steamlinkAttach

from .web import WebApp
from .db import DB

from .util import (
	getargs,
	loadconfig,
	createconfig,
	daemonize,
	check_pid,
	write_pid,
)

from .testdata import TestData

from .const import (
	PROJECT_PACKAGE_NAME,
	__version__,
	DEFAULT_CONFIG_FILE,
)


class GracefulExit(SystemExit):
	code = 1

class GracefulRestart(SystemExit):
	code = 2

def raise_graceful_restart():
	logger.info("restarting")
	raise GracefulRestart()

def raise_graceful_exit():
	logger.info("shutdown")
	raise GracefulExit()

home = str(pathlib.Path.home())	

		

#
# Default config
# specified config file (if any) will get merged in here
DEFAULT_CONF = OrderedDict({
 	'general': OrderedDict({
		'mqtt_broker': 'mqtt_broker',
		'ping_timeout': 30,
		'working_dir': home + '/.steamlink',
		'max_log_records': 1000
	}),
	'Steam': OrderedDict({
		'id': 0,
		'name': 'sample1',
		'description': 'SteamLink Sample',
		'namespace': '/sl',
		'autocreate': False,
	}),
	'tests': OrderedDict({
		'test1': OrderedDict({
			'desc': 'Test1',
		    'startwait': 2,
		    'meshes': 3,
		    'nodes': 12,
		    'packets': 1000,
		    'pkt_delay': 0.0,
		})
	}),
	'mqtt': OrderedDict({
		'clientid': None,
		'username': None,
		'password': None,
		'server': '127.0.0.1',
		'port': 1883,
		'ssl_certificate': None,

		'prefix': 'SteamLink',
		'data': 'data',
		'control': 'control',
		'public_control': 'SteamLink/pub/%s/control',
		'public_data': 'SteamLink/pub/%s/data',
	}),

	'console': OrderedDict({
		'host': '0.0.0.0',
		'port': 5050,
		'shutdown_timeout': 10,        # seconds to wait for web server shutdown
		'namespace': '/sl',
		'minupdinterval': 1.0,
		'index': "",           # root page
        'ssl_certificate': None,
        'ssl_key': None,
		'repo_name': 'steamlink',
		'repo_owner': 'steamlink',
		'repo_key': 'SteamLinkIsIoT',
		'repo_ci_command': None,
	}),
	'mqtt_broker':  OrderedDict({
		'listeners': OrderedDict({
			'default': OrderedDict({
				'type': 'tcp',
				'bind': '127.0.0.1:1883',
			}),
			'ws-mqtt': OrderedDict({
				'bind': '127.0.0.1:8080',
				'type': 'ws',
				'max_connections': 10,
			}),
		}),
		'sys_interval': 10,
		'auth': OrderedDict({
			'allow-anonymous': True,
			'#password-file': os.path.join(os.path.dirname(os.path.realpath(__file__)), 'passwd'),
			'plugins': [
				'#auth_file',
				'auth_anonymous',
			]
		})
	}),
	'DB':  OrderedDict({
		'db_filename': home + '/.steamlink/steamlink.db',
	})
})


#
# Main
#
def steamlink_main(cl_args, conf):
	global daemon, conf_working_dir, logq

	""" start steamlink """

	if conf['mqtt']['clientid'] is None:
		conf['mqtt']['clientid'] = "clie"+"%04i" % int(random.random() * 10000)

	conf_general = conf['general']
	conf_working_dir = conf_general['working_dir']
	if not os.path.exists(conf_working_dir):
		os.mkdir(conf_working_dir)

	conf_console = conf['console']
	conf_steam = conf['Steam']
	conf_db = conf['DB']

	# Daemon functions
	if cl_args.pid_file:
		check_pid(cl_args.pid_file)

	daemon = cl_args.daemon
	if cl_args.daemon:
		if cl_args.verbose:
			print("continuing in background")
		daemonize()
	if cl_args.pid_file:
		write_pid(cl_args.pid_file)

	# N.B. no asyncio before daemon!
	aioloop = asyncio.get_event_loop()
	if logging.DBG >= 2:
		aioloop.set_debug(enabled=True)

	try:
		aioloop.add_signal_handler(signal.SIGHUP, raise_graceful_restart)
		aioloop.add_signal_handler(signal.SIGINT, raise_graceful_exit)
		aioloop.add_signal_handler(signal.SIGTERM, raise_graceful_exit)
	except NotImplementedError:  # pragma: no cover
		logger.error("main: failed to trap signals")

	broker_c = conf_general.get('mqtt_broker', None)
	if broker_c is None:
		conf_broker = None
	else:
		conf_broker = conf.get(broker_c, None)
		if conf_broker is None:
			logger.error("mqtt broker section '%s' does not exist", broker_c)
			sys.exit(1)
	
	conf_mqtt = conf['mqtt']

	namespace = conf_steam['namespace']


	# confugure subsystems
	coros = []

	LogQ._table = DictTable(LogQ, keyfield="name", index = OrderedDict())
	logq = LogQ(conf, aioloop)
	coros.append(logq.start())

	msghandler = logging.StreamHandler(logq)
	QFORMAT = '%(levelname)s %(module)s %(message)s'
	qfmt = logging.Formatter(QFORMAT)
	msghandler.setFormatter(qfmt)
	msghandler.setLevel(logging.INFO)
	logger.addHandler(msghandler)

	logger.info("%s version %s" % (PROJECT_PACKAGE_NAME, __version__))

	if conf_broker is not None:
		logger.debug("startup: create MQTT Broker")
		mqtt_broker = Mqtt_Broker(conf_broker, loop=aioloop)
		logger.debug("startup: start MQTT Broker")
		try:
			aioloop.run_until_complete(mqtt_broker.start())
		except hbmqtt.broker.BrokerException as e:
#			logger.error("mqtt: broker start failed: %s", e)
			sys.exit(1)

	logger.debug("startup: create Mqtt")
	mqtt = Mqtt(conf_mqtt)

	aioloop.run_until_complete(mqtt.start())

	ping_timeout = conf_general['ping_timeout']

	ll = logging.getLogger('AsyncServer')
	# use different logging level socketio/engineio modules
	ll.setLevel(logging.WARN)
	logger.debug("startup: create socketio")
	sio = socketio.AsyncServer(
		logger = ll,
		async_mode = 'aiohttp',
#		cors_allowed_origins =  "http://localhost:* http://127.0.0.1:*",
#		cors_credentials = True,
		ping_timeout = ping_timeout,
		engineio_logger = ll,
	)

	logger.debug("startup: open DB")
	db = DB(conf_db, loop=aioloop)

	logger.debug("startup: create WebApp")
	webapp = None
	webapp = WebApp(namespace, sio, conf_console, loop=aioloop)

	linkageAttach(webapp, db)
	steamlinkAttach(mqtt, db)

	logger.debug("startup: start db")
	aioloop.run_until_complete(db.start())

	logger.debug("startup: start webapp")
	aioloop.run_until_complete(webapp.start())
	
	SteamSetup()
	steam = Steam(conf_steam)
	set_steam_root(steam)

	if cl_args.testdata:
		testconfigs = conf['tests']
		logger.debug("startup: create TestData")
		if not cl_args.testdata in testconfigs:
			logger.error("testdata: no section '%s' in config file", cl_args.testdata)
			sys.exit(1)
		TestTask = TestData(testconfigs[cl_args.testdata], aioloop)
		logger.debug("startup: starting TestData")
		coros.append(TestTask.start())
	else:
		TestTask = None

	ll = logging.getLogger('asyncio_socket')
	ll.setLevel(logging.WARN)

	coros.append(steam.start())

	coros.append(webapp.qstart())
	logger.debug("startup: starting coros")
	restart = False
	try:
		aioloop.run_until_complete(asyncio.gather(
			*coros
			))
		aioloop.run_forever()
	except (GracefulRestart):
		logger.info("restarting")
		restart = True
	except (GracefulExit, KeyboardInterrupt):
		logger.info("shutting down")
	except hbmqtt.errors.NoDataException as e:
		logger.notice("coros run_until: hbmqtt.errors.NoDataException: %s", e)

	# Shutdown
	webapp.stop()
	aioloop.run_until_complete(db.stop())
	if TestTask:
		logger.debug("stopping TestTask")
		TestTask.stop()

	aioloop.run_until_complete(mqtt.stop())
	if not conf_broker is None:
		aioloop.run_until_complete(mqtt_broker.stop())

	logger.info("done")
	return restart


#
# Main
#

daemon = False
def steamlink_command():
	global daemon
	cl_args = getargs()

	if cl_args.version:
		print("%s version %s" % (PROJECT_PACKAGE_NAME, __version__))
		return 0

	try:
		loglevel = getattr(logging, cl_args.loglevel.upper())
	except Exception as e:
		loglevel = None

	if loglevel is None:
		loglevel = logging.DEBUG if cl_args.debug > 0 else logging.INFO

	FORMAT = '%(asctime)-15s: %(levelname)s %(module)s %(message)s'
#	logging.basicConfig(format=FORMAT)
	logger.setLevel(loglevel)
	if cl_args.logfile is not None:
		logging.console = False
		handler = logging.handlers.RotatingFileHandler(
              cl_args.logfile, maxBytes=10**7, backupCount=3)
	else:
		logging.console = True
		handler = logging.StreamHandler()

	fmt = logging.Formatter(FORMAT)
	handler.setFormatter(fmt)
	logger.addHandler(handler)

	DBG = cl_args.debug
	logging.DBG = DBG

	DBGK = cl_args.debugkey
	logging.DBGK = DBGK

	if DBG >= 3:
		logger.info("DBG: logging all warnings")
		import warnings
		warnings.simplefilter("always")
		logging.captureWarnings(True)

	if cl_args.config is None:
		conff = home + "/" + DEFAULT_CONFIG_FILE
	else:
		conff = cl_args.config

	# create config  if -C
	if cl_args.createconfig:
		rc = createconfig(conff, DEFAULT_CONF)
		return(rc)

	# load config
	conf = loadconfig(DEFAULT_CONF, conff)

	try:
		restart = steamlink_main(cl_args, conf)
		rc = 0
#	except SystemExit as e:
#		rc  = e
#		pass
	except:
		restart = False
		rc = 127
		exc_type, exc_value, exc_traceback = sys.exc_info()
		tb = traceback.format_exception(exc_type, exc_value, exc_traceback)
		if not daemon:
			for l in tb:
				print(l.rstrip('\n'))
		else:
			syslog.syslog(syslog.LOG_ERR, ' main failed')
			for l in tb:
				syslog.syslog(syslog.LOG_ERR, ' -> %s' % l.rstrip('\n'))
				logger.error(' -> %s', l.rstrip('\n'))
	if restart:
		os.execve(sys.argv[0], sys.argv, os.environ)
		print("os.execve returned")
	return rc

if __name__ == "main":
	rc = steamlink_command()
	sys.exit(rc)
